"""Parser for teal instructions.

Each teal instruction is represented as a class. Parsing teal instructions
is creating the correct Instruction subclass instance and setting the fields
of it given the source code line.

Each teal instruction in the source code start with a fixed prefix. A instruction
might contain additional immediate values which need to be handled case by case.
Instruction parser works by first creating a list where each item has the fixed prefix
for the instruction and a callable(function) that will handle parsing of immediate
arguments specific to that instruction. All the callables take a string containing
all the immediate arguments and return the instruction object with the parsed immediates.

In teal contracts, a source line might contain comments, indentation, additional to the
instruction. Teal comments in the same line as the instruction will be stored in the
instruction object. Indentation, comments without a instruction, empty lines are all
ignored by the instruction parser.

Attributes:
    parser_rules (List[Tuple[str, Callable[[str], Instruction]]]):
        List of tuple each containing the instruction prefix and a callable.
        Callable handles the parsing of immediate arguments of the instruction and
        creation of instruction object.

Exceptions:
    ParseError: Exception class to represent errors raised while parsing
        invalid or incorrectly formatted instructions.
"""

from typing import Optional, Callable, Tuple, List
import base64

from tealer.teal.instructions import instructions
from tealer.teal.instructions.instructions import Instruction
from tealer.teal.instructions.parse_acct_params_field import parse_acct_params_field
from tealer.teal.instructions.parse_app_params_field import parse_app_params_field
from tealer.teal.instructions.parse_asset_holding_field import parse_asset_holding_field
from tealer.teal.instructions.parse_asset_params_field import parse_asset_params_field
from tealer.teal.instructions.parse_global_field import parse_global_field
from tealer.teal.instructions.parse_transaction_field import parse_transaction_field


class ParseError(Exception):
    """Exception class to represent instruction parsing errors."""


def handle_gtxn(x: str, itxn: bool = False) -> Instruction:
    """Parse gtxn instruction.

    Args:
        x: proper string representation of gtxn instruction immediate
            arguments.
        itxn: True if the instruction is gitxn

    Returns:
        Gtxn instruction object created after parsing immediate
        arguments.
    """

    split = x.split(" ")
    idx = _parse_int(split[0])
    tx_field = parse_transaction_field(" ".join(split[1:]), False)
    if itxn:
        return instructions.Gitxn(idx, tx_field)
    return instructions.Gtxn(idx, tx_field)


def handle_gtxna(x: str, itxn: bool = False) -> Instruction:
    """Parse gtxna instruction.

    Args:
        x: proper string representation of gtxna instruction immediate
            arguments.
        itxn: True if the instruction is gitxna

    Returns:
        Gtxna instruction object created after parsing immediate
        arguments.
    """

    split = x.split(" ")
    idx = _parse_int(split[0])
    tx_field = parse_transaction_field(" ".join(split[1:]), False)
    if itxn:
        return instructions.Gitxna(idx, tx_field)
    return instructions.Gtxna(idx, tx_field)


def handle_gtxnas(x: str, itxn: bool = False) -> Instruction:
    """Parse gtxnas instruction.

    Args:
        x: proper string representation of gtxnas instruction
            immediate arguments.
        itxn: True if the instruction is gitxnas

    Returns:
        Gtxnas instruction object created after parsing immediate
        arguments.
    """

    args = x.split(" ")
    idx = _parse_int(args[0])
    tx_field = parse_transaction_field(args[1], True)
    if itxn:
        return instructions.Gitxnas(idx, tx_field)
    return instructions.Gtxnas(idx, tx_field)


def _parse_int(x: str) -> int:
    """Parse teal integers.

    Teal supports three formats to write integers: hex, octal and
    decimal. hexadecimal numbers start with the prefix 0x and octal
    numbers have prefix 0.

    Args:
        x: string representation of the teal integer.

    Returns:
        python integer equal to the value represented by the given
        teal integer.
    """

    if x.startswith("0x"):
        return int(x[2:], 16)
    if x.startswith("0"):
        return int(x, 8)
    return int(x)


def _split_instruction_into_tokens(line: str) -> List[str]:
    """Split given instruction into tokens.

    Teal instructions are divided into tokens. Tokens in teal are
    comments, string literals or sequence of non-space characters.

    Args:
        line (str): string representation of teal instruction.

    Returns:
        List[str]: List of tokens.

    Examples:
        >>> _split_instruction_into_tokens('a "string" // comment')
        ['a', '"string"', '// comment']
        >>> _split_instruction_into_tokens('a "s tring //" // "comment"')
        ['a', '"s tring //"', '// "comment"']

    Raises:
        ParseError: raises ParseError if string literals are not
            properly enclosed in the line.
    """

    fields: List[str] = []
    line = line.strip()
    i = 0
    start = i
    while i < len(line):
        if line[i].isspace():
            if start != i:
                fields.append(line[start:i].strip())
            i += 1
            start = i
        elif line[i] == '"':
            if start != i:
                i += 1
                continue
            i += 1
            while i < len(line):
                if line[i] == '"' and line[i - 1] != "\\":
                    fields.append(line[start : i + 1])
                    i += 1
                    start = i
                    break
                i += 1
            else:
                raise ParseError(f"missing closing qoute {line}")
        elif line[i : i + 2] == "//":
            fields.append(line[i:])
            return fields
        else:
            i += 1
    if start < len(line):
        fields.append(line[start:].strip())
    return fields


def _b64_decode(s: str) -> str:
    """Return hex representation of base64 encoded string.

    Args:
        s: base64 encoded string.

    Returns:
        hex representation of the given string with the prefix "0x"
    """

    return "0x" + base64.b64decode(s + "=" * (-len(s) % 4)).hex()


def _b32_decode(s: str) -> str:
    """Return hex representation of base32 encoded string.

    Args:
        s: base32 encoded string.

    Returns:
        hex representation of the given string with the prefix "0x"
    """

    return "0x" + base64.b32decode(s + "=" * (-len(s) % 8)).hex()


def _parse_byte_arguments(fields: List[str]) -> List[str]:
    """Parse multiple representations of teal byte literals.

    Parses byte literals and convert all base encoded strings to hex
    encoding. currently parses following formats
    base64 AAAA..., b64 AAAA..., base64(AAAA...), b64(AAAA...),
    base32 AAAA..., b32 AAAA..., base32(AAAA...), b32(AAAA...),
    0x0123..., "literal".

    Args:
        fields (List[str]): list of space separated fields(tokens).

    Returns:
        List[str]: list of byte arguments, all either encoded in
            hex or as string literals.

    Examples:
        >>> _parse_byte_arguments(['base64', 'AA', 'base64(AA)', 'b64(AA)'])
        ['0x00', '0x00', '0x00']
        >>> _parse_byte_arguments(['0x00', '"str"', 'base32', 'AA'])
        ['0x00', '"str"', '0x00']

    Raises:
        ParseError: raises ParseError if the byte arguments are not correctly
            encoded.
    """

    arguments: List[str] = []
    error: bool = False
    i: int = 0
    while i < len(fields) and not error:
        if fields[i] == "base64" or fields[i] == "b64":
            if len(fields) <= i + 1:
                error = True
                break
            arguments.append(_b64_decode(fields[i + 1]))
            i += 2
        elif fields[i] == "base32" or fields[i] == "b32":
            if len(fields) <= i + 1:
                error = True
                break
            arguments.append(_b32_decode(fields[i + 1]))
            i += 2
        elif fields[i].startswith("base64(") or fields[i].startswith("b64("):
            if fields[i][-1] != ")":
                error = True
                break
            data = fields[i].split("(")[1][:-1]
            arguments.append(_b64_decode(data))
            i += 1
        elif fields[i].startswith("base32(") or fields[i].startswith("b32("):
            if fields[i][-1] != ")":
                error = True
                break
            data = fields[i].split("(")[1][:-1]
            arguments.append(_b32_decode(data))
            i += 1
        elif fields[i].startswith("0x") or fields[i].startswith('"'):
            arguments.append(fields[i])
            i += 1
        else:
            error = True
            i += 1
            break
    if error:
        line = " ".join(fields)
        raise ParseError(f"incorrect byte format {line}")
    return arguments


# return true if the given string is a teal integer.
_is_int: Callable[[str], bool] = lambda x: x.startswith("0x") or x.isdigit()


# Order in the parser_rules is important
parser_rules: List[Tuple[str, Callable[[str], Instruction]]] = [
    ("#pragma version ", lambda x: instructions.Pragma(_parse_int(x))),
    ("err", lambda _x: instructions.Err()),
    ("assert", lambda _x: instructions.Assert()),
    ("int ", lambda x: instructions.Int(_parse_int(x) if _is_int(x) else x)),
    ("pushints ", lambda x: instructions.PushInts(list(map(_parse_int, x.split(" "))))),
    ("pushint ", lambda x: instructions.PushInt(_parse_int(x) if _is_int(x) else x)),
    ("txn ", lambda x: instructions.Txn(parse_transaction_field(x, False))),
    ("txna ", lambda x: instructions.Txna(parse_transaction_field(x, False))),
    ("gtxn ", lambda x: handle_gtxn(x)),
    ("gtxna ", lambda x: handle_gtxna(x)),
    ("gtxns ", lambda x: instructions.Gtxns(parse_transaction_field(x, False))),
    ("gtxnsa ", lambda x: instructions.Gtxnsa(parse_transaction_field(x, False))),
    ("load ", lambda x: instructions.Load(_parse_int(x))),
    ("store ", lambda x: instructions.Store(_parse_int(x))),
    (
        "gload ",
        lambda x: instructions.Gload(_parse_int(x.split(" ")[0]), _parse_int(x.split(" ")[1])),
    ),
    ("gloads ", lambda x: instructions.Gloads(_parse_int(x))),
    ("gloadss", lambda _x: instructions.Gloadss()),
    ("gaid ", lambda x: instructions.Gaid(_parse_int(x))),
    ("gaids", lambda x: instructions.Gaids()),
    ("loads", lambda _x: instructions.Loads()),
    ("stores", lambda _x: instructions.Stores()),
    ("dig ", lambda x: instructions.Dig(_parse_int(x))),
    ("swap", lambda _x: instructions.Swap()),
    ("getbit", lambda _x: instructions.GetBit()),
    ("setbit", lambda _x: instructions.SetBit()),
    ("getbyte", lambda _x: instructions.GetByte()),
    ("setbyte", lambda _x: instructions.SetByte()),
    (
        "extract ",
        lambda x: instructions.Extract(_parse_int(x.split(" ")[0]), _parse_int(x.split(" ")[1])),
    ),
    ("extract3", lambda _x: instructions.Extract3()),
    ("extract_uint16", lambda _x: instructions.Extract_uint16()),
    ("extract_uint32", lambda _x: instructions.Extract_uint32()),
    ("extract_uint64", lambda _x: instructions.Extract_uint64()),
    ("sha256", lambda _x: instructions.Sha256()),
    ("sha512_256", lambda _x: instructions.Sha512_256()),
    ("keccak256", lambda _x: instructions.Keccak256()),
    ("ed25519verify_bare", lambda _x: instructions.Ed25519verify_bare()),
    ("ed25519verify", lambda _x: instructions.Ed25519verify()),
    ("ecdsa_verify", lambda x: instructions.Ecdsa_verify(x)),
    ("ecdsa_pk_decompress", lambda x: instructions.Ecdsa_pk_decompress(x)),
    ("ecdsa_pk_recover", lambda x: instructions.Ecdsa_pk_recover(x)),
    ("global ", lambda x: instructions.Global(parse_global_field(x))),
    ("dupn ", lambda x: instructions.Dupn(_parse_int(x))),
    ("dup2", lambda _x: instructions.Dup2()),
    ("dup", lambda _x: instructions.Dup()),
    ("select", lambda _x: instructions.Select()),
    ("cover", lambda x: instructions.Cover(_parse_int(x))),
    ("uncover", lambda x: instructions.Uncover(_parse_int(x))),
    ("concat", lambda _x: instructions.Concat()),
    ("b ", lambda x: instructions.B(x)),
    ("bz ", lambda x: instructions.BZ(x)),
    ("bnz ", lambda x: instructions.BNZ(x)),
    ("return", lambda x: instructions.Return()),
    ("callsub ", lambda x: instructions.Callsub(x)),
    ("retsub", lambda x: instructions.Retsub()),
    ("app_global_get_ex", lambda x: instructions.AppGlobalGetEx()),
    ("app_global_get", lambda x: instructions.AppGlobalGet()),
    ("app_global_put", lambda x: instructions.AppGlobalPut()),
    ("app_global_del", lambda x: instructions.AppGlobalDel()),
    ("app_local_get_ex", lambda x: instructions.AppLocalGetEx()),
    ("app_local_get", lambda x: instructions.AppLocalGet()),
    ("app_local_put", lambda x: instructions.AppLocalPut()),
    ("app_local_del", lambda x: instructions.AppLocalDel()),
    ("app_opted_in", lambda x: instructions.AppOptedIn()),
    ("balance", lambda x: instructions.Balance()),
    ("min_balance", lambda x: instructions.MinBalance()),
    ("asset_holding_get ", lambda x: instructions.AssetHoldingGet(parse_asset_holding_field(x))),
    ("asset_params_get ", lambda x: instructions.AssetParamsGet(parse_asset_params_field(x))),
    ("app_params_get ", lambda x: instructions.AppParamsGet(parse_app_params_field(x))),
    ("acct_params_get ", lambda x: instructions.AcctParamsGet(parse_acct_params_field(x))),
    ("%", lambda x: instructions.Modulo()),
    ("!=", lambda x: instructions.Neq()),
    ("!", lambda x: instructions.Not()),
    ("==", lambda x: instructions.Eq()),
    ("&&", lambda x: instructions.And()),
    ("&", lambda x: instructions.BitwiseAnd()),
    ("||", lambda x: instructions.Or()),
    ("|", lambda x: instructions.BitwiseOr()),
    ("+", lambda x: instructions.Add()),
    ("-", lambda x: instructions.Sub()),
    ("/", lambda x: instructions.Div()),
    ("*", lambda x: instructions.Mul()),
    (">=", lambda x: instructions.GreaterE()),
    (">", lambda x: instructions.Greater()),
    ("<=", lambda x: instructions.LessE()),
    ("<", lambda x: instructions.Less()),
    ("^", lambda x: instructions.BitwiseXor()),
    ("~", lambda x: instructions.BitwiseInvert()),
    ("bitlen", lambda _x: instructions.BitLen()),
    ("b%", lambda x: instructions.BModulo()),
    ("b!=", lambda x: instructions.BNeq()),
    ("b==", lambda x: instructions.BEq()),
    ("b&", lambda x: instructions.BBitwiseAnd()),
    ("b|", lambda x: instructions.BBitwiseOr()),
    ("b+", lambda x: instructions.BAdd()),
    ("b-", lambda x: instructions.BSubtract()),
    ("b/", lambda x: instructions.BDiv()),
    ("b*", lambda x: instructions.BMul()),
    ("b>=", lambda x: instructions.BGreaterE()),
    ("b>", lambda x: instructions.BGreater()),
    ("b<=", lambda x: instructions.BLessE()),
    ("b<", lambda x: instructions.BLess()),
    ("b^", lambda x: instructions.BBitwiseXor()),
    ("b~", lambda x: instructions.BBitwiseInvert()),
    ("bzero", lambda x: instructions.BZero()),
    ("bsqrt", lambda x: instructions.BSqrt()),
    ("log", lambda _x: instructions.Log()),
    ("itxn_begin", lambda _x: instructions.Itxn_begin()),
    ("itxn_next", lambda _x: instructions.Itxn_next()),
    ("itxn_field ", lambda x: instructions.Itxn_field(parse_transaction_field(x, True))),
    ("itxn_submit", lambda _x: instructions.Itxn_submit()),
    ("itxn ", lambda x: instructions.Itxn(parse_transaction_field(x, False))),
    ("itxna ", lambda x: instructions.Itxna(parse_transaction_field(x, False))),
    ("gitxn ", lambda x: handle_gtxn(x, itxn=True)),
    ("gitxna ", lambda x: handle_gtxna(x, itxn=True)),
    ("gitxnas ", lambda x: handle_gtxnas(x, itxn=True)),
    ("txnas ", lambda x: instructions.Txnas(parse_transaction_field(x, True))),
    ("itxnas ", lambda x: instructions.Itxnas(parse_transaction_field(x, True))),
    ("gtxnas ", lambda x: handle_gtxnas(x)),
    ("gtxnsas ", lambda x: instructions.Gtxnsas(parse_transaction_field(x, True))),
    ("args", lambda _x: instructions.Args()),
    ("itob", lambda x: instructions.Itob()),
    ("btoi", lambda x: instructions.Btoi()),
    ("popn ", lambda x: instructions.Popn(_parse_int(x))),
    ("pop", lambda x: instructions.Pop()),
    ("addr ", lambda x: instructions.Addr(x)),
    ("mulw", lambda x: instructions.Mulw()),
    ("addw", lambda x: instructions.Addw()),
    ("divmodw", lambda x: instructions.Divmodw()),
    ("divw", lambda x: instructions.Divw()),
    ("expw", lambda x: instructions.Expw()),
    ("exp", lambda x: instructions.Exp()),
    ("shl", lambda x: instructions.Shl()),
    ("shr", lambda x: instructions.Shr()),
    ("sqrt", lambda x: instructions.Sqrt()),
    ("intcblock", lambda x: instructions.Intcblock(list(map(_parse_int, x.strip().split())))),
    ("intc ", lambda x: instructions.Intc(_parse_int(x))),
    ("intc_0", lambda x: instructions.Intc0()),
    ("intc_1", lambda x: instructions.Intc1()),
    ("intc_2", lambda x: instructions.Intc2()),
    ("intc_3", lambda x: instructions.Intc3()),
    ("bytec ", lambda x: instructions.Bytec(_parse_int(x))),
    ("bytec_0", lambda x: instructions.Bytec0()),
    ("bytec_1", lambda x: instructions.Bytec1()),
    ("bytec_2", lambda x: instructions.Bytec2()),
    ("bytec_3", lambda x: instructions.Bytec3()),
    ("arg ", lambda x: instructions.Arg(_parse_int(x))),
    ("arg_0", lambda x: instructions.Arg0()),
    ("arg_1", lambda x: instructions.Arg1()),
    ("arg_2", lambda x: instructions.Arg2()),
    ("arg_3", lambda x: instructions.Arg3()),
    ("len", lambda x: instructions.Len()),
    (
        "substring ",
        lambda x: instructions.Substring(_parse_int(x.split(" ")[0]), _parse_int(x.split(" ")[1])),
    ),
    ("substring3", lambda x: instructions.Substring3()),
    ("replace2 ", lambda x: instructions.Replace2(_parse_int(x))),
    ("replace3", lambda _x: instructions.Replace3()),
    ("replace", lambda x: instructions.Replace(None if x == "" else _parse_int(x))),
    ("base64_decode ", lambda x: instructions.Base64_decode(x)),
    ("json_ref ", lambda x: instructions.Json_ref(x)),
    ("sha3_256", lambda _x: instructions.Sha3_256()),
    ("vrf_verify ", lambda x: instructions.Vrf_verify(x)),
    ("block ", lambda x: instructions.Block(x)),
    ("bury ", lambda x: instructions.Bury(_parse_int(x))),
    (
        "proto ",
        lambda x: instructions.Proto(_parse_int(x.split(" ")[0]), _parse_int(x.split(" ")[1])),
    ),
    ("frame_dig ", lambda x: instructions.FrameDig(_parse_int(x))),
    ("frame_bury ", lambda x: instructions.FrameBury(_parse_int(x))),
    ("switch ", lambda x: instructions.Switch(x.split(" "))),
    ("match ", lambda x: instructions.Match(x.split(" "))),
    ("box_create", lambda x: instructions.BoxCreate()),
    ("box_extract", lambda x: instructions.BoxExtract()),
    ("box_replace", lambda x: instructions.BoxReplace()),
    ("box_del", lambda x: instructions.BoxDel()),
    ("box_len", lambda x: instructions.BoxLen()),
    ("box_get", lambda x: instructions.BoxGet()),
    ("box_put", lambda x: instructions.BoxPut()),
]


def parse_line(line: str) -> Optional[instructions.Instruction]:
    """Parse line into a teal instruction.

    This might return None instead of an instance of Instruction subclass
    if the line is empty, contains only comments, etc.

    Args:
        line: Teal source code line to parse.

    Returns:
        returns instance of Instruction subclass corresponding to the given
        string representation if :line: contains a valid representation or
        else returns None.

    Raises:
        ParseError: raises ParseError if the instruction present in the given
            line is not correctly formatted and if is not valid.
    """
    if not line.strip():
        return None

    source_code_line = line
    fields = _split_instruction_into_tokens(line)
    comment = ""
    if fields[-1].startswith("//"):
        comment = fields[-1]
        fields = fields[:-1]

    if not fields:
        return None

    ins: Optional[Instruction] = None
    if fields[0][-1] == ":":
        if len(fields) != 1:
            raise ParseError(f"incorrect format of label: {line}")
        ins = instructions.Label(fields[0][:-1])

    if fields[0] == "byte" or fields[0] == "pushbytes" or fields[0] == "method":
        imm: List[str] = _parse_byte_arguments(fields[1:])
        if len(imm) != 1:
            raise ParseError(f"{fields[0]} expects exactly one argument: {line}")
        ins = {
            "byte": instructions.Byte,
            "pushbytes": instructions.PushBytes,
            "method": instructions.Method,
        }[fields[0]](imm[0])

    if fields[0] == "bytecblock":
        imm = _parse_byte_arguments(fields[1:])
        ins = instructions.Bytecblock(imm)

    if fields[0] == "pushbytess":
        imm = _parse_byte_arguments(fields[1:])
        ins = instructions.PushBytess(imm)

    if ins is not None:
        ins.comment = comment
        ins.source_code = source_code_line
        return ins

    line = " ".join(fields)

    f: Callable[[str], Instruction]
    for key, f in parser_rules:
        if line.startswith(key):
            ins = f(line[len(key) :].strip())
            ins.comment = comment
            ins.source_code = source_code_line
            return ins

    # line is checked to not empty at the start of the function.
    print(f'Not found instruction: "{line}"')
    ins = instructions.UnsupportedInstruction(line)
    ins.source_code = source_code_line
    ins.comment = ins.comment
    return ins



from typing import Tuple, List, Type
import pytest

from tealer.teal.basic_blocks import BasicBlock
from tealer.detectors.abstract_detector import AbstractDetector
from tealer.utils.command_line.common import init_tealer_from_single_contract
from tealer.utils.output import ExecutionPaths, InstructionsOutput

from tests.detectors.groupsize import missing_group_size_tests
from tests.detectors.fee_check import missing_fee_check_tests, new_missing_fee_tests
from tests.detectors.can_close_account import can_close_account_tests, new_can_close_account_tests
from tests.detectors.can_close_asset import can_close_asset_tests, new_can_close_asset_tests
from tests.detectors.can_delete import can_delete_tests, new_can_delete_tests
from tests.detectors.can_update import can_update_tests, new_can_update_tests
from tests.detectors.optimizations.constant_gtxn import constant_gtxn
from tests.detectors.optimizations.sender_access import sender_access
from tests.detectors.optimizations.self_access import self_access
from tests.detectors.rekeyto import missing_rekeyto_tests, new_missing_rekeyto_tests
from tests.detectors.subroutine_patterns import subroutine_patterns_tests

from tests.utils import cmp_cfg


ALL_TESTS: List[Tuple[str, Type[AbstractDetector], List[List[BasicBlock]]]] = [
    *missing_group_size_tests,
    *missing_fee_check_tests,
    *can_close_account_tests,
    *can_close_asset_tests,
    *can_delete_tests,
    *can_update_tests,
    *missing_rekeyto_tests,
]


@pytest.mark.parametrize("test", ALL_TESTS)  # type: ignore
def test_detectors(test: Tuple[str, Type[AbstractDetector], List[List[BasicBlock]]]) -> None:
    code, detector, expected_paths = test
    tealer = init_tealer_from_single_contract(code.strip(), "test")
    tealer.register_detector(detector)
    result = tealer.run_detectors()[0][0]

    if isinstance(result, ExecutionPaths):
        assert len(result.paths) == len(expected_paths)
        for path, ex_path in zip(result.paths, expected_paths):
            assert cmp_cfg(path, ex_path)

    else:
        # Not implemented yet
        assert False


ALL_NEW_TESTS: List[Tuple[str, Type[AbstractDetector], List[List[int]]]] = [
    *new_can_close_account_tests,
    *new_missing_rekeyto_tests,
    *new_can_close_asset_tests,
    *new_can_update_tests,
    *new_can_delete_tests,
    *new_missing_fee_tests,
    *subroutine_patterns_tests,
]


@pytest.mark.parametrize("test", ALL_NEW_TESTS)  # type: ignore
def test_just_detectors(test: Tuple[str, Type[AbstractDetector], List[List[int]]]) -> None:
    code, detector, expected_paths = test
    tealer = init_tealer_from_single_contract(code.strip(), "test")
    tealer.register_detector(detector)
    result = tealer.run_detectors()[0][0]

    if isinstance(result, ExecutionPaths):
        print(f"count: result = {len(result.paths)}, expected = {len(expected_paths)}")
        assert len(result.paths) == len(expected_paths)
        for path, expected_path in zip(result.paths, expected_paths):
            print(f"path length: result = {len(path)}, expected = {len(expected_path)}")
            print(f"path ids: result = {path}, expected = {expected_path}")
            assert len(path) == len(expected_path)
            for bi, expected_idx in zip(path, expected_path):
                assert bi.idx == expected_idx

    else:
        # Not implemented yet
        assert False


# This test is for the detectors using InstructionsOutput
# This type return only a list of instructions
# So we compare the expected lines of the source code
# source code -> detector -> lust(expected lines)
TESTS_INSTRUCTIONS_OUTPUT: List[Tuple[str, Type[AbstractDetector], List[List[int]]]] = [
    *constant_gtxn,
    *sender_access,
    *self_access,
]


@pytest.mark.parametrize("test", TESTS_INSTRUCTIONS_OUTPUT)  # type: ignore
def test_instructions_output_detectors(
    test: Tuple[str, Type[AbstractDetector], List[List[int]]]
) -> None:
    code, detector, expected_instruction_lines = test
    tealer = init_tealer_from_single_contract(code.strip(), "test")
    tealer.register_detector(detector)
    result = tealer.run_detectors()[0][0]

    if isinstance(result, InstructionsOutput):
        print(
            f"count: result = {len(result.instructions)}, expected = {len(expected_instruction_lines)}"
        )

        actual_instruction_lines = [
            [ins.line for ins in instructions] for instructions in result.instructions
        ]

        assert actual_instruction_lines == expected_instruction_lines

    else:
        # Not implemented yet
        assert False



# pylint: skip-file
# mypy: ignore-errors
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import

from tealer.detectors.all_detectors import IsUpdatable

router = Router(
    name="Test",
    bare_calls=BareCallActions(
        no_op=OnCompleteAction(action=Approve(), call_config=CallConfig.ALL),
    ),
)


@router.method(update_application=CallConfig.ALL)
def update_application() -> Expr:
    return Seq(
        [
            # IsUpdatable should report this path and will do so.
            # Assert(Global.creator_address() == Txn.sender()),
            Return(),
        ]
    )


working_approval_program, _, _ = router.compile_program(
    version=7,
)

false_negative_router = Router(
    name="FalseNegative",
    bare_calls=BareCallActions(
        no_op=OnCompleteAction(action=Approve(), call_config=CallConfig.CREATE),
    ),
)


@Subroutine(TealType.none)
def temp_sub() -> Expr:
    return Pop(Bytes("Call this subroutine twice"))


@false_negative_router.method(update_application=CallConfig.CALL)
def update_application_2() -> Expr:
    return Seq(
        [
            # IsUpdatable should report this path and but does not report it.
            # reason:
            #   1. When temp_sub is called for the first time, it's blocks are added
            #       to the path.
            #   2. And To avoid loops, detectors check that basic block they are processing
            #       is not already present in the `path`.
            #   3. When temp_sub is called for the second time, detectors think that it is a back edge/loop
            #       because `temp_sub` basic blocks are already added to the path on the first call.
            #
            #   TLDR; detectors do not report vulnerable paths, if there atleast 2 calls to the same subroutine in the path.
            # Above bug is fixed and tests should pass
            temp_sub(),
            temp_sub(),
            Return(),
        ]
    )


false_negative_program, _, _ = false_negative_router.compile_program(
    version=7,
)

working_approval_program_paths = [[0, 1, 3, 8, 4]]
false_negative_program_paths = [[0, 1, 3, 9, 8, 10, 8, 11, 4]]


method_returns_using_approve = Router(
    name="MethodReturnsUsingApprove",
    bare_calls=BareCallActions(
        no_op=OnCompleteAction(action=Approve(), call_config=CallConfig.CREATE),
    ),
)


@method_returns_using_approve.method(update_application=CallConfig.CALL)
def update_application_3() -> Expr:
    return Approve()


method_returns_using_approve_program, _, _ = method_returns_using_approve.compile_program(
    version=7,
)

method_returns_using_approve_program_teal = """
#pragma version 7       // B0
txn NumAppArgs
int 0
==
bnz main_l4

txna ApplicationArgs 0          // B1
method "update_application_3()void"
==
bnz main_l3

err             // B2

main_l3:            // B3
txn OnCompletion
int UpdateApplication
==
txn ApplicationID
int 0
!=
&&
assert
callsub updateapplication3_0

int 1       // B4
return

main_l4:    // B5
txn OnCompletion
int NoOp
==
bnz main_l6

err         // B6

main_l6:        // B7
txn ApplicationID
int 0
==
assert
int 1
return

// update_application_3
updateapplication3_0:   // B8
int 1
return
"""

# Each of router methods have a corresponding subroutine in the compiled code.
# `updateapplication3_0` subroutine for update_application_3() method in above example.
# PyTeal implicitly adds instructions `int 1; return;` after the call to the subroutine of the method.
# See block B4 in above code.
# PyTeal adds those instructions even if the subroutine uses `return` instead of `retsub`.
# This is a problem for Tealer. Because, B4 is considered as a return point to `callsub` instruction.
# B4 does not have any predecessors and execution is not reachable.
# During the TransactionFieldAnalysis, information of return point block is used to calculate information for callsub block.
# And because B4 is not reachable, it will have empty set for possible values. In backward propagation, information from B4
# is intersected with information of B3 resulting in empty set of possible values for B3 as well. This propagates across the CFG
# and blocks will have incorrect information.

# This bug is now fixed.

method_returns_using_approve_program_paths = [[0, 1, 3, 8]]

method_returns_using_approve_2 = Router(
    name="MethodReturnsUsingApprove",
    bare_calls=BareCallActions(
        no_op=OnCompleteAction(action=Approve(), call_config=CallConfig.CREATE),
    ),
)


@method_returns_using_approve_2.method(update_application=CallConfig.CALL)
def update_application_3() -> Expr:
    # return in one path of the subroutine.
    # And retsub in another path of the subroutine.
    return If(Txn.application_args[0] == Bytes("Update"), Approve(), Return())


method_returns_using_approve_2_program, _, _ = method_returns_using_approve_2.compile_program(
    version=7
)

method_returns_using_approve_2_program_paths = [
    [0, 1, 3, 8, 9, 4],
    [0, 1, 3, 8, 10],
]

multiple_calls_to_subroutine_tests = [
    (working_approval_program, IsUpdatable, working_approval_program_paths),
    (false_negative_program, IsUpdatable, false_negative_program_paths),
    (method_returns_using_approve_program, IsUpdatable, method_returns_using_approve_program_paths),
    (
        method_returns_using_approve_2_program,
        IsUpdatable,
        method_returns_using_approve_2_program_paths,
    ),
]

if __name__ == "__main__":
    print(working_approval_program)
    print(false_negative_program)
    print(method_returns_using_approve_program)
    print(method_returns_using_approve_2_program)



"""Defines class to represent basic blocks of teal contract.

Tealer works on Control Flow Graph(CFG) of the contract for
analysis. CFG is represented as a list of basic blocks in tealer.
This module defines the BasicBlock class for representing basic blocks,
parts of contract code that have only single entry and exit point of
execution. BasicBlock class contains methods, properties used to store,
retrieve previous and next basic blocks. This previous and next basic blocks
of a basic block represent edges coming into and coming out of it
i.e they represents edges in the Control Flow Graph(CFG).

Classes:
    BasicBlock: Class to represent basic blocks of teal contract.
"""

from typing import List, Optional, TYPE_CHECKING

from tealer.teal.instructions.instructions import Instruction, Callsub, Retsub
from tealer.exceptions import TealerException

if TYPE_CHECKING:
    from tealer.teal.teal import Teal
    from tealer.teal.subroutine import Subroutine


class BasicBlock:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Class to represent basic blocks of the teal contract.

    A basic block is a sequence of instructions with a single entry
    and single exit point. The instructions in a basic block always execute
    in the same sequence from the entry point to the exit point if at all
    executed. This class is used to store instructions that belong to
    a single BasicBlock in teal contracts.
    """

    def __init__(self) -> None:
        self._instructions: List[Instruction] = []
        self._prev: List[BasicBlock] = []
        self._next: List[BasicBlock] = []
        self._idx: int = 0
        self._teal: Optional["Teal"] = None
        self._tealer_comments: List[str] = []
        self._subroutine: Optional["Subroutine"] = None

    def add_instruction(self, instruction: Instruction) -> None:
        """Append instruction to this basic block.

        Args:
            instruction: intruction to add to the basic block.
        """

        self._instructions.append(instruction)

    @property
    def instructions(self) -> List[Instruction]:
        """List of instructions part of this basic block.

        Returns:
            list of basic block instructions.
        """
        return self._instructions

    @property
    def entry_instr(self) -> Instruction:
        """Entry(first) instruction of this basic block.

        Returns:
            The entry instruction of the basic block.
        """
        return self._instructions[0]

    @property
    def exit_instr(self) -> Instruction:
        """Exit(Last) instruction of this basic block.

        Returns:
            The exit instruction of the basic block.
        """
        return self._instructions[-1]

    def add_prev(self, prev_bb: "BasicBlock") -> None:
        """Add basic block that may execute just before this basic block.

        Executing a basic block means that the execution is passed to
        the entry instruction of the basic block. Previous basic blocks are
        basic blocks whose exit instruction might execute before the entry
        instruction of this basic block i.e exit instruction of previous
        basic block is previous instruction of entry instruction of this
        basic block.

        Args:
            prev_bb: basic block to add to the list of previous
                basic blocks of this basic block.
        """

        self._prev.append(prev_bb)

    def add_next(self, next_bb: "BasicBlock") -> None:
        """Add basic block that may execute right after this instruction.

        A basic block is considered as next basic block if the execution may be
        passed from this basic block to that basic block. All the basic blocks
        whose entry instruction is next instruction of exit instruction of this
        basic block are next basic blocks.

        Args:
            next_bb: basic block to add to the list of next basic blocks
                of this basic block.
        """

        self._next.append(next_bb)

    @property
    def prev(self) -> List["BasicBlock"]:
        """List of previous basic blocks to this basic block.

        Returns:
            list of previous instructions that might have been executed before this block.
        """
        return self._prev

    @property
    def next(self) -> List["BasicBlock"]:
        """List of next basic blocks to this basic block.

        Returns:
            list of next basic blocks that might be executed after this block.
        """
        return self._next

    @property
    def idx(self) -> int:
        """Index of this basic block when ordered by line number of entry instruction.

        Returns:
            index of this basic block in list of all basic blocks.
        """
        return self._idx

    @idx.setter
    def idx(self, i: int) -> None:
        self._idx = i

    @property
    def cost(self) -> int:
        """cost of executing all instructions in this basic block

        Returns:
            Returns the OpcodeCost of executing all instructions in the block.
        """
        return sum(ins.cost for ins in self.instructions)

    @property
    def teal(self) -> Optional["Teal"]:
        """Teal instance of the contract this basic block belongs to.

        Returns:
            Returns the contract of this basic block.
        """
        return self._teal

    @teal.setter
    def teal(self, teal_instance: "Teal") -> None:
        self._teal = teal_instance

    @property
    def subroutine(self) -> "Subroutine":
        """Subroutine instrance of the subroutine this basic block belongs to.

        Returns:
            Returns the subroutine this block belongs to.

        Raises:
            TealerException: raises error if subroutine is not set during parsing.
        """
        if self._subroutine is None:
            raise TealerException(f"subroutine of B{self._idx} is not initialized")
        return self._subroutine

    @subroutine.setter
    def subroutine(self, subroutine_instance: "Subroutine") -> None:
        self._subroutine = subroutine_instance

    @property
    def is_callsub_block(self) -> bool:
        """Return True if the block calls a subroutine.

        Returns:
            Returns True if this block has a callsub instruction.
        """
        return isinstance(self.exit_instr, Callsub)

    @property
    def called_subroutine(self) -> "Subroutine":
        """Return the subroutine called by this subroutine.

        Returns:
            Returns the subroutine called by the callsub instruction in this block.

        Raises:
            TealerException: if this block is not a callsub_block.
        """
        if not isinstance(self.exit_instr, Callsub):
            raise TealerException("called subroutine of a non callsub block is accessed")
        return self.exit_instr.called_subroutine

    @property
    def sub_return_point(self) -> Optional["BasicBlock"]:
        """Return the return point block of this block.

        Returns:
            The subroutine return_point block of this(callsub) block.
            Returns None if this callsub block does not have a return point block. This happens when
            callsub instruction is the last instruction in the contract and The called subroutine always exits
            the program.

        Raises:
            TealerException: if this block is not a callsub_block.
        """
        if not self.is_callsub_block:
            raise TealerException("sub_return_point block of a non callsub block is accessed")
        return self.next[0] if self.next else None

    @property
    def is_sub_return_point(self) -> bool:
        """Return True if this block is executed after the subroutine i.e next block of callsub_block

        Returns:
            Returns True if this block is executed a subroutine: The next block of a block with callsub
            instruction. Otherwise, returns False.
        """
        for bi in self.prev:
            if bi.is_callsub_block:
                return True
        return False

    @property
    def callsub_block(self) -> "BasicBlock":
        """Return the callsub_block which calls the subroutine. This block is the return point block.

        Returns:
            Return the callsub_block which calls the subroutine that returns the execution to this
            basic block.

        Raises:
            TealerException: if this block is not a sub_return_point block.
        """
        for bi in self.prev:
            if bi.is_callsub_block:
                return bi
        raise TealerException("callsub_block of a non sub_return_point block is accessed")

    @property
    def is_retsub_block(self) -> bool:
        return isinstance(self.exit_instr, Retsub)

    @property
    def tealer_comments(self) -> List[str]:
        """Additional comments added by tealer for each basic block in the output CFG.

        Returns:
            Returns the additional comments added by tealer for this basic block.
        """
        return self._tealer_comments

    @tealer_comments.setter
    def tealer_comments(self, comments: List[str]) -> None:
        self._tealer_comments = comments

    def __str__(self) -> str:
        ret = ""
        for ins in self._instructions:
            ret += f"{ins.line}: {ins}\n"
        return ret

    def __repr__(self) -> str:
        return f"B{self.idx}"



"""Detect paths missing validations on sender field AND allows to update the application."""

from typing import List, TYPE_CHECKING, Tuple

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.detectors.utils import (
    detect_missing_tx_field_validations_group,
    detect_missing_tx_field_validations_group_complete,
)
from tealer.utils.teal_enums import TealerTransactionType
from tealer.utils.output import ExecutionPaths

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.utils.output import ListOutput
    from tealer.teal.context.block_transaction_context import BlockTransactionContext
    from tealer.teal.teal import Teal


class AnyoneCanUpdate(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find execution paths missing validations on sender field AND allows to update the application.

    Stateful smart contracts(application) can be updated in algorand. If the
    application transaction of type UpdateApplication is approved by the application,
    then the application's code will be updated with the new code.

    This detector tries to find execution paths for which
        - OnCompletion can be UpdateApplication And
        - Transaction sender can be any address.
    """

    NAME = "unprotected-updatable"
    DESCRIPTION = "Unprotected Upgradable Applications"
    TYPE = DetectorType.STATEFULL

    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = "https://github.com/crytic/tealer/wiki/Detector-Documentation#unprotected-updatable-application"
    WIKI_TITLE = "Unprotected Upgradable Application"
    WIKI_DESCRIPTION = (
        "Application can be updated by anyone. "
        "More at [building-secure-contracts/not-so-smart-contracts/algorand/access_controls]"
        "(https://github.com/crytic/building-secure-contracts/tree/master/not-so-smart-contracts/algorand/access_controls)."
    )
    WIKI_EXPLOIT_SCENARIO = """
```py
@router.method(update_application=CallConfig.CALL)
def update_application() -> Expr:
    return Approve()
```

Eve updates the application by calling `update_application` method and steals all of its assets.
"""

    WIKI_RECOMMENDATION = """
- Avoid upgradable applications.
- Add access controls to the vulnerable method.
"""

    def detect(self) -> "ListOutput":
        """Detect paths missing validations on sender field AND allows to update the application.

        Returns:
            ExecutionPaths instance containing the list of vulnerable execution
            paths along with name, check, impact, confidence and other detector
            information.
        """

        def checks_field(block_ctx: "BlockTransactionContext") -> bool:
            # return False if Txn Type can be UpdateApplication AND sender can be any address.
            # return True otherwise.
            return not (
                TealerTransactionType.ApplUpdateApplication in block_ctx.transaction_types
                and block_ctx.sender.any_addr
            )

        # there should be a better to decide which function to call ??
        if self.tealer.output_group:
            # mypy complains if the value is returned directly. Uesd the second suggestion mentioned here:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#variance
            return list(
                detect_missing_tx_field_validations_group_complete(self.tealer, self, checks_field)
            )

        output: List[
            Tuple["Teal", List[List["BasicBlock"]]]
        ] = detect_missing_tx_field_validations_group(self.tealer, checks_field)
        detector_output: "ListOutput" = []
        for contract, vulnerable_paths in output:
            detector_output.append(ExecutionPaths(contract, self, vulnerable_paths))

        return detector_output



from typing import TYPE_CHECKING, List, Optional

from tealer.exceptions import TealerException
from tealer.teal.instructions.instructions import Retsub

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.teal import Teal


class Subroutine:  # pylint: disable=too-many-instance-attributes
    """Represent a Teal subroutine.

    Main entry point code, code that is not part of any subroutine, is represented as a separate subroutine.
    """

    def __init__(self, name: str, entry: "BasicBlock", blocks: List["BasicBlock"]) -> None:
        self._name = name
        self._entry = entry
        self._blocks = blocks
        self._exit_blocks = [
            b for b in blocks if len(b.next) == 0 or isinstance(b.exit_instr, Retsub)
        ]
        self._contract: Optional["Teal"] = None
        self._caller_callsub_blocks: List["BasicBlock"] = []
        self._return_point_blocks: List["BasicBlock"] = []

    @property
    def name(self) -> str:
        """Name of the subroutine

        Returns:
            Returns the name of the subroutine. The name of the CFG with non-subroutine blocks is "__main__"
        """
        return self._name

    @property
    def entry(self) -> "BasicBlock":
        """Entry block of the subroutine.

        Returns:
            entry block of the subroutine.
        """
        return self._entry

    @property
    def exit_blocks(self) -> List["BasicBlock"]:
        """Exit blocks of the subroutine.

        Returns:
            Returns the exit blocks of the subroutine. Exit blocks include blocks which return from the subroutine(retsub)
            and also blocks which exit from the contract.
        """
        return self._exit_blocks

    @property
    def blocks(self) -> List["BasicBlock"]:
        """List of all basic blocks of the subroutine.

        Returns:
            Returns list of all basic blocks that are part of this subroutine.
        """
        return self._blocks

    @property
    def contract(self) -> "Teal":
        """The Teal contract this subroutine is a part of

        Returns:
            Returns the contract.

        Raises:
            TealerException: Raises error if contract is not set during parsing.
        """
        if self._contract is None:
            raise TealerException(f"Contract of subroutine {self._name} is not set")
        return self._contract

    @contract.setter
    def contract(self, contract_obj: "Teal") -> None:
        self._contract = contract_obj

    @property
    def caller_blocks(self) -> List["BasicBlock"]:
        """BasicBlock with callsub instructions which call this subroutine.

        Returns:
            List of caller callsub basic blocks. This blocks have "callsub {self.name}"
            as exit instruction.
        """
        return self._caller_callsub_blocks

    @caller_blocks.setter
    def caller_blocks(self, caller_callsub_blocks: List["BasicBlock"]) -> None:
        """Set caller_blocks and return point blocks

        Args:
            caller_callsub_blocks: basic blocks with callsub instruction calling this subroutine.
        """
        self._caller_callsub_blocks = caller_callsub_blocks
        self._return_point_blocks = [
            bi.next[0] for bi in caller_callsub_blocks if len(bi.next) == 1
        ]

    @property
    def return_point_blocks(self) -> List["BasicBlock"]:
        return self._return_point_blocks

    @property
    def retsub_blocks(self) -> List["BasicBlock"]:
        """List of basic blocks of the subroutine containing `Retsub`.

        Returns:
            Returns the list of retsub blocks. This blocks return the execution from the
            subroutine to the caller.
        """
        return [b for b in self._exit_blocks if isinstance(b.exit_instr, Retsub)]

    @property
    def called_subroutines(self) -> List["Subroutine"]:
        """List of subroutines called by this subroutine.

        Returns:
            Returns a list of subroutines called by the subroutine.
        """
        return list(set(bi.called_subroutine for bi in self._blocks if bi.is_callsub_block))



# type: ignore[unreachable]
import sys
from typing import Tuple, List, Type

import pytest

from tealer.detectors.abstract_detector import AbstractDetector
from tealer.utils.command_line.common import init_tealer_from_single_contract
from tealer.utils.output import ExecutionPaths

if not sys.version_info >= (3, 10):
    pytest.skip(reason="PyTeal based tests require python >= 3.10", allow_module_level=True)

# Place import statements after the version check
from tests.detectors.router_with_assembled_constants import (  # pylint: disable=wrong-import-position
    router_with_assembled_constants,
)
from tests.detectors.pyteal_can_close import (  # pylint: disable=wrong-import-position
    txn_type_based_tests,
)
from tests.detectors.pyteal_group_size import (  # pylint: disable=wrong-import-position
    group_size_tests_pyteal,
)

from tests.detectors.pyteal_subroutine_recursion import (  # pylint: disable=wrong-import-position
    subroutine_recursion_patterns_tests,
)

from tests.detectors.multiple_calls_to_subroutine import (  # pylint: disable=wrong-import-position
    multiple_calls_to_subroutine_tests,
)

TESTS: List[Tuple[str, Type[AbstractDetector], List[List[int]]]] = [
    *router_with_assembled_constants,
    *txn_type_based_tests,
    *group_size_tests_pyteal,
    *subroutine_recursion_patterns_tests,
    *multiple_calls_to_subroutine_tests,
]


@pytest.mark.parametrize("test", TESTS)  # type: ignore
def test_detectors_using_pyteal(test: Tuple[str, Type[AbstractDetector], List[List[int]]]) -> None:
    code, detector, expected_paths = test
    tealer = init_tealer_from_single_contract(code.strip(), "test")
    tealer.register_detector(detector)
    result = tealer.run_detectors()[0]
    if not isinstance(result, ExecutionPaths):
        result = result[0]

    # for bi in teal.bbs:
    #     print(
    #         bi,
    #         bi.idx,
    #         bi.transaction_context.transaction_types,
    #         bi.transaction_context.group_indices,
    #     )
    #     print(
    #         bi.transaction_context.gtxn_context(0).transaction_types,
    #         bi.transaction_context.group_indices,
    #     )
    #     print(
    #         bi.transaction_context.gtxn_context(1).transaction_types,
    #         bi.transaction_context.group_indices,
    #     )
    print("result paths =", result.paths)
    print("expected paths =", expected_paths)
    assert len(result.paths) == len(expected_paths)
    for path, expected_path in zip(result.paths, expected_paths):
        assert len(path) == len(expected_path)
        for bi, expected_idx in zip(path, expected_path):
            assert bi.idx == expected_idx



"""Detector for finding execution paths missing AssetCloseTo check."""

from typing import List, TYPE_CHECKING, Tuple

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.detectors.utils import (
    detect_missing_tx_field_validations_group,
    detect_missing_tx_field_validations_group_complete,
)
from tealer.utils.teal_enums import TealerTransactionType
from tealer.utils.output import ExecutionPaths
from tealer.utils.teal_enums import TransactionType

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.utils.output import ListOutput
    from tealer.teal.context.block_transaction_context import BlockTransactionContext
    from tealer.teal.teal import Teal


class CanCloseAsset(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find execution paths missing AssetCloseTo check.

    In algorand, A transaction can close out the asset balance of a contract
    account and transfer all asset balance to the specified address. If the
    AssetCloseTo field of a transaction is set to an address and that transaction
    is approved by the contract then all the contract asset balance will be
    transferred to address in AssetCloseTo field.

    This detector tries to find execution paths that approve the algorand
    transaction("return 1") and doesn't check the AssetCloseTo field.
    """

    NAME = "can-close-asset"
    DESCRIPTION = "Missing AssetCloseTo Field Validation"
    TYPE = DetectorType.STATELESS

    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = "https://github.com/crytic/tealer/wiki/Detector-Documentation#missing-assetcloseto-field-validation"
    WIKI_TITLE = "Missing AssetCloseTo Field Validation"
    WIKI_DESCRIPTION = (
        "LogicSig does not validate `AssetCloseTo` field."
        " Attacker can submit a transaction with `AssetCloseTo` field set to their address and steal account's assets."
        " More at [building-secure-contracts/not-so-smart-contracts/algorand/closing_asset]"
        "(https://github.com/crytic/building-secure-contracts/tree/master/not-so-smart-contracts/algorand/closing_asset)"
    )
    WIKI_EXPLOIT_SCENARIO = """
```py
def withdraw(...) -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.type_enum() == TxnType.AssetTransfer,
                    Txn.first_valid() % period == Int(0),
                    Txn.last_valid() == Txn.first_valid() + duration,
                    Txn.asset_receiver() == receiver,
                    Txn.asset_amount() == amount,
                    Txn.first_valid() < timeout,
                )
            ),
            Approve(),
        ]
    )
```

Alice signs the logic-sig to allow recurring payments to Bob in USDC.\
 Eve uses the logic-sig and submits a valid transaction with `AssetCloseTo` field set to her address.\
 Eve steals Alice's USDC balance.
"""

    WIKI_RECOMMENDATION = """
Validate `AssetCloseTo` field in the LogicSig.
"""

    def detect(self) -> "ListOutput":
        """Detect execution paths with missing AssetCloseTo check.

        Returns:
            ExecutionPaths instance containing the list of vulnerable execution
            paths along with name, check, impact, confidence and other detector
            information.
        """

        def checks_field(block_ctx: "BlockTransactionContext") -> bool:
            # return False if AssetCloseTo field can have any address.
            # return True if AssetCloseTo should have some address or zero address
            # AssetCloseTo field can only be set for Asset Transafer type transactions.
            return not (
                block_ctx.assetcloseto.any_addr
                and TealerTransactionType.Axfer in block_ctx.transaction_types
            )

        # there should be a better to decide which function to call ??
        if self.tealer.output_group:
            # mypy complains if the value is returned directly. Uesd the second suggestion mentioned here:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#variance
            return list(
                detect_missing_tx_field_validations_group_complete(
                    self.tealer,
                    self,
                    checks_field,
                    [TransactionType.Any, TransactionType.Unknown, TransactionType.Axfer],
                )
            )

        output: List[
            Tuple["Teal", List[List["BasicBlock"]]]
        ] = detect_missing_tx_field_validations_group(self.tealer, checks_field)
        detector_output: "ListOutput" = []
        for contract, vulnerable_paths in output:
            detector_output.append(ExecutionPaths(contract, self, vulnerable_paths))

        return detector_output



from typing import Union

from tealer.utils.comparable_enum import ComparableEnum


class TealerTransactionType(ComparableEnum):
    Pay = 0x10
    KeyReg = 0x20
    Acfg = 0x30
    Axfer = 0x40
    Afrz = 0x50
    ApplNoOp = 0x60
    ApplOptIn = 0x61
    ApplCloseOut = 0x62
    ApplClearState = 0x63
    ApplUpdateApplication = 0x64
    ApplDeleteApplication = 0x65
    ApplCreation = 0x66
    Appl = 0x67

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name


class TransactionType(ComparableEnum):
    Invalid = 0x0
    Pay = 0x1
    KeyReg = 0x02
    Acfg = 0x3
    Axfer = 0x4
    Afrz = 0x5
    Appl = 0x6

    Any = 0x7
    Unknown = 0x8


def transaction_type_from_txt(txn_type: str) -> TransactionType:
    return {
        "pay": TransactionType.Pay,
        "keyreg": TransactionType.KeyReg,
        "acfg": TransactionType.Acfg,
        "axfer": TransactionType.Axfer,
        "afrz": TransactionType.Afrz,
        "appl": TransactionType.Appl,
        "txn": TransactionType.Any,
    }[txn_type]


class TransactionOnCompletion(ComparableEnum):
    NoOp = 0x0
    OptIn = 0x1
    CloseOut = 0x2
    ClearState = 0x3
    UpdateApplication = 0x4
    DeleteApplication = 0x5


ALL_TRANSACTION_TYPES = (
    TealerTransactionType.Pay,
    TealerTransactionType.KeyReg,
    TealerTransactionType.Acfg,
    TealerTransactionType.Axfer,
    TealerTransactionType.Appl,
    TealerTransactionType.ApplNoOp,
    TealerTransactionType.ApplOptIn,
    TealerTransactionType.ApplCloseOut,
    TealerTransactionType.ApplClearState,
    TealerTransactionType.ApplUpdateApplication,
    TealerTransactionType.ApplDeleteApplication,
    TealerTransactionType.ApplCreation,
)

APPLICATION_TRANSACTION_TYPES = (
    TealerTransactionType.ApplNoOp,
    TealerTransactionType.ApplOptIn,
    TealerTransactionType.ApplCloseOut,
    TealerTransactionType.ApplClearState,
    TealerTransactionType.ApplUpdateApplication,
    TealerTransactionType.ApplDeleteApplication,
    TealerTransactionType.ApplCreation,
)
TYPEENUM_TRANSACTION_TYPES = (
    TealerTransactionType.Pay,
    TealerTransactionType.KeyReg,
    TealerTransactionType.Acfg,
    TealerTransactionType.Axfer,
    TealerTransactionType.Afrz,
    TealerTransactionType.Appl,
)


def oncompletion_to_tealer_type(value: Union[str, int]) -> "TealerTransactionType":
    ENUM_NAMES_TO_INT = {
        "NoOp": 0,
        "OptIn": 1,
        "CloseOut": 2,
        "ClearState": 3,
        "UpdateApplication": 4,
        "DeleteApplication": 5,
    }
    INT_TO_TYPE = {
        0: TealerTransactionType.ApplNoOp,
        1: TealerTransactionType.ApplOptIn,
        2: TealerTransactionType.ApplCloseOut,
        3: TealerTransactionType.ApplClearState,
        4: TealerTransactionType.ApplUpdateApplication,
        5: TealerTransactionType.ApplDeleteApplication,
    }

    if not isinstance(value, int):
        value = ENUM_NAMES_TO_INT[value]

    return INT_TO_TYPE[value]


def transaction_type_to_tealer_type(value: Union[str, int]) -> "TealerTransactionType":
    ENUM_NAMES_TO_INT = {
        "pay": 1,
        "keyreg": 2,
        "acfg": 3,
        "axfer": 4,
        "afrz": 5,
        "appl": 6,
    }
    INT_TO_TYPE = {
        1: TealerTransactionType.Pay,
        2: TealerTransactionType.KeyReg,
        3: TealerTransactionType.Acfg,
        4: TealerTransactionType.Axfer,
        5: TealerTransactionType.Afrz,
        6: TealerTransactionType.Appl,
    }

    if not isinstance(value, int):
        value = ENUM_NAMES_TO_INT[value]

    return INT_TO_TYPE[value]


class ExecutionMode(ComparableEnum):
    """Execution mode of the instruction/function/contract.

    Stateless: Contract do not have access to persistent storage, block information, etc.
        All stateless contracts are LogicSigs.
    Stateful: Contracts have access to persistent storage and are referred to as Applications.
        A stateful contract can be an ApprovalProgram or a ClearStateProgram.
    Any: Contract can be executed in any mode, either stateless or stateful.
    """

    STATELESS = 0
    STATEFUL = 1
    ANY = 2

    def __str__(self) -> str:
        return self.name.title()


class ContractType(ComparableEnum):
    """Type of the contract.

    LogicSig: Contract is used to sign the transactions.
    ApprovalProgram: Contract is deployed onto the chain and will be executed for calls with
        on-complete value of NoOp, OptIn, CloseOut, UpdateApplication and DeleteApplication.
    ClearStateProgram: Contract is deployed onto the chain and will be executed for calls with
        on-complete value of ClearState.
    """

    LogicSig = 0
    ApprovalProgram = 1
    ClearStateProgram = 2

    Unknown = 99

    def __str__(self) -> str:
        return self.name


def contract_type_from_txt(contract_type: str) -> ContractType:
    return {
        "ApprovalProgram": ContractType.ApprovalProgram,
        "ClearStateProgram": ContractType.ClearStateProgram,
        "LogicSig": ContractType.LogicSig,
    }[contract_type]



from typing import Tuple, List
import pytest

from tealer.utils.command_line.common import init_tealer_from_single_contract
from tests.utils import order_basic_blocks


BRANCHING_ADDR = """
#pragma version 2
txn Receiver
addr 6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM
==
bz wrongreceiver
txn Fee
int 10000
<
bz highfee
txn CloseRemainderTo
global ZeroAddress
==
bz closing_account
txn AssetCloseTo
global ZeroAddress
==
bz closing_asset
global GroupSize
int 1
==
bz unexpected_group_size
int 1
return
wrongreceiver:
highfee:
closing_account:
closing_asset:
unexpected_group_size:
    err
"""

rekeyto_any = [True] * 6 + [False] * 5
rekeyto_none = [False] * 6 + [True] * 5
rekeyto: List[List[str]] = [[] for _ in range(11)]
rekeyto_values = (rekeyto_any, rekeyto_none, rekeyto)

closeto_any = [False] * 11
closeto_none = [True] * 11
closeto: List[List[str]] = [[] for _ in range(11)]
closeto_values = (closeto_any, closeto_none, closeto)

assetcloseto_any = [False] * 11
assetcloseto_none = [True] * 11
assetcloseto: List[List[str]] = [[] for _ in range(11)]
assetcloseto_values = (assetcloseto_any, assetcloseto_none, assetcloseto)

BRANCHING_ADDR_VALUES = [
    rekeyto_values,
    closeto_values,
    assetcloseto_values,
]

BRANCHING_ADDR_GTXN = """
#pragma version 2
txn Receiver
addr 6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM
==
bz wrongreceiver
txn Fee
int 10000
<
bz highfee
gtxn 0 CloseRemainderTo
global ZeroAddress
==
bz closing_account
gtxn 0 AssetCloseTo
addr AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEVAL4QAJS7JHB4
==
bz closing_asset
global GroupSize
int 1
==
bz unexpected_group_size
int 1
return
wrongreceiver:
highfee:
closing_account:
closing_asset:
unexpected_group_size:
    err
"""

ALL_TESTS_TXN = [
    (BRANCHING_ADDR, BRANCHING_ADDR_VALUES, -1),
    (BRANCHING_ADDR_GTXN, BRANCHING_ADDR_VALUES, 0),
]


@pytest.mark.parametrize("test", ALL_TESTS_TXN)  # type: ignore
def test_addr_fields(  # pylint: disable=too-many-locals
    test: Tuple[str, List[Tuple[List[bool], List[bool], List[List[str]]]], int]
) -> None:
    code, values, idx = test
    tealer = init_tealer_from_single_contract(code.strip(), "test")
    function = tealer.contracts["test"].functions["test"]
    ex_rekeyto_any, ex_rekeyto_none, ex_rekeyto = values[0]
    ex_closeto_any, ex_closeto_none, ex_closeto = values[1]
    ex_assetcloseto_any, ex_assetcloseto_none, ex_assetcloseto = values[2]

    bbs = order_basic_blocks(function.blocks)
    for b in bbs:
        if idx == -1:
            ctx = function.transaction_context(b)
        else:
            ctx = function.transaction_context(b).gtxn_context(idx)

        assert ctx.rekeyto.any_addr == ex_rekeyto_any[b.idx]
        assert ctx.rekeyto.no_addr == ex_rekeyto_none[b.idx]
        assert set(ctx.rekeyto.possible_addr) == set(ex_rekeyto[b.idx])

        assert ctx.closeto.any_addr == ex_closeto_any[b.idx]
        assert ctx.closeto.no_addr == ex_closeto_none[b.idx]
        assert set(ctx.closeto.possible_addr) == set(ex_closeto[b.idx])

        assert ctx.assetcloseto.any_addr == ex_assetcloseto_any[b.idx]
        assert ctx.assetcloseto.no_addr == ex_assetcloseto_none[b.idx]
        assert set(ctx.assetcloseto.possible_addr) == set(ex_assetcloseto[b.idx])



import logging
from typing import List, TYPE_CHECKING, Callable, Tuple, Optional

from tealer.utils.analyses import next_blocks_global, leaf_block_global
from tealer.detectors.abstract_detector import DetectorType
from tealer.utils.output import GroupTransactionOutput
from tealer.utils.teal_enums import TransactionType

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.teal import Teal
    from tealer.tealer import Tealer
    from tealer.teal.functions import Function
    from tealer.teal.subroutine import Subroutine
    from tealer.teal.context.block_transaction_context import BlockTransactionContext
    from tealer.detectors.abstract_detector import AbstractDetector

logger_detectors = logging.getLogger("Detectors")
logging.basicConfig(level=logging.DEBUG)


def validated_in_block(
    block: "BasicBlock",
    function: "Function",
    checks_field: Callable[["BlockTransactionContext"], bool],
    absolute_index: Optional[int] = None,
) -> bool:
    """
    A transaction field can be validated by accessing using instructions
    - txn {field}
    - gtxn {current_transaction_index} {field}

    `block.transaction_context` is information from `txn {field}` instructions.
    `block.transaction_context.gtxn_context({index})` is information from `gtxn {index} {field}` instructions.
    block.transaction_context.gtxn_context(i) represents values `transaction field` have when it's index is `i`.
    Note that, block.transaction_context.gtxn_context(i) does not represent possible values of any transaction at index
    `i`. It only represents possible values of current txn given that txn index is `i`.

    Args:
        block: A basic block of the CFG
        function: The function
        checks_field: A function which given a block context, should return True if the target field
            cannot have the vulnerable value or else False.
            e.g: For is_updatable detector, vulnerable value is `UpdateApplication`.
            if `block_ctx.transaction_types` can have `UpdateApplication`, this method will
            return false or else returns True.
        absolute_index: The index of the transaction (given in group config or computed)

    Returns:
        returns True if the field(s) is validated in this block or else False
    """
    # print(str(block), "\n", block, function.transaction_context(block).transaction_types)
    # if field is checked using `txn {field}`, return true
    if checks_field(function.transaction_context(block)):
        return True
    # if absolute index is given, check if field was checked by accessing it using gtxn(s) instructions`
    if absolute_index is not None:
        if checks_field(function.transaction_context(block).gtxn_context(absolute_index)):
            return True
        return False
    # for each possible {index} the transaction can have, check if field is checked using `gtxn {index} {field}`
    for i in function.transaction_context(block).group_indices:
        if not checks_field(function.transaction_context(block).gtxn_context(i)):
            return False
    # the field is checked for all possible indices.
    return True


def detect_missing_tx_field_validations(
    function: "Function",
    checks_field: Callable[["BlockTransactionContext"], bool],
    satisfies_report_condition: Callable[[List["BasicBlock"]], bool] = lambda _x: True,
) -> List[List["BasicBlock"]]:
    """search for execution paths lacking validations on transaction fields.

    `tealer.analyses.dataflow` calculates possible values a transaction field can have to successfully complete execution.
    Information is computed at block level. Each block's context(`BlockTransactionContext`)
    will have information on possible values for a transaction field such that
        - Execution reaches to the first instruction of block `B`
        - Execution reaches exit instruction of block `B` without failing.

    The goal of current detectors is to find if a given transaction field(s) can have a certain vulnerable value(s).
    e.g: is_updatable detector checks if the transaction can be UpdateApplication type transaction. Here, the vulnerable value for
    transaction type is "UpdateApplication". if "UpdateApplication" is a possible value then the contract is considered vulnerable and is reported.

    In order to find whether the given contract is vulnerable, it is sufficient to check the leaf blocks: whether transaction field can
    have target value and execution reaches end of a leaf block.
    However, The current output format requires "a execution path" from entry block to leaf block. And when a leaf block is vulnerable,
    it does **not** mean that all execution paths ending at that leaf block are vulnerable. It only means that atleast one path ending at
    that leaf block is vulnerable.

    For this reason, a traversal over the CFG is performed to enumerate execution paths. And for each block in the path it is checked
    whether transaction field(s) can have the target vulnerable value.

    Args:
        function: The function being checked
        checks_field: Given a block context, should return True if the target field cannot have the vulnerable value
            or else False.
        satisfies_report_condition: Given a path, should return True if the "path" satifies the vulnerable condition.

        Example for checks_field
            e.g: For is_updatable detector, vulnerable value is `UpdateApplication`.
            if `block_ctx.transaction_types` can have `UpdateApplication`, this method will
            return false or else returns True.

        Example for satisfies_report_condition
            The group-size detectors reports path lacking GroupSize check. However, it only reports the path
            if an absolute index is used by one of the blocks in the path. That dectector can use this function
            to decide on whether to report or not.
            Other detectors can also use additional constraints that are based on entire vulnerable path before
            reporting.

    Returns:
        Returns a list of vulnerable execution paths: none of the blocks in the path check the fields.
    """

    def search_paths(
        bb: "BasicBlock",
        current_path: List["BasicBlock"],
        paths_without_check: List[List["BasicBlock"]],
        current_call_stack: List[Tuple[Optional["BasicBlock"], "Subroutine"]],
        current_subroutine_executed: List[List["BasicBlock"]],
    ) -> None:
        """
        Traverse the CFG, generate possible execution paths and report vulnerable execution paths.

        Execution paths are generated for the global CFG, where callsub blocks are connected to the subroutine's
        entry block.

        Args:
            bb: current basic block.
            current_path: List of basic blocks in the execution path from the contract's entry block to
                the current basic block bb.
            paths_without_check: List of execution paths considered vulnerable out of the generated execution paths
                upto now.
            current_call_stack: list of callsub blocks and called subroutine along the current path.
                e.g current_callsub_blocks = [(Bi, S1), (Bj, S2), (Bk, S3), ...]
                => 1. Bi, Bj, Bk, .. all end with callsub instruction.
                   2. Bi called subroutine S1. S1 contains block Bj which called subroutine S2.
                        S2 contains Bk which calls S3.
                        S1 calls S2, S2 calls S3, ...
            current_subroutine_executed: list of subroutine basic blocks that were executed in the path.
                e.g if call stack is [(Bi, S1), (Bj, S2), (Bk, S3)]
                1. current_subroutine_executed[0] contains basic blocks of S1 that were executed before call to S2.
                2. current_subroutine_executed[1] contains basic blocks of S2 that were executed before call to S3.
                3. current_subroutine_executed[2] contains basic blocks of S3 that were executed upto this call.

                Main purpose of this argument is to identify loops. A loop contains a back edge to a previously executed block.
                if presence of loop is checked using `return True if bb in current_path else False`, detector misses paths that
                does not contain a loop.
                Mainly, when a subroutine is called twice in a path, the second call is treated as a back edge because the subroutine's blocks
                will be present in the current path (added in the first call.)
                e.g
                    ```
                    callsub Sa  // B0
                    callsub Sa  // B1
                    int 1       // B2
                    return
                    Sa:         // B3
                        callsub Sb
                        retsub      // B4
                    Sb:         // B5
                        retsub
                    ```
                before executing B1, current_path = [B0, B3, B5, B4], bb = B1
                after executing B1, current_path = [B0, B3, B5, B4, B1], bb = B3

                B3 is already present in the current_path and so is treated as a loop even though it's not.

                In order to identify loops properly, basic blocks that are executed as part of the current subroutines are tracked.
                For above example,
                    At the start of the function
                        bb = B0, current_path = [], current_call_stack = [(None, Main)], current_subroutine_executed = [[]]

                    At the end of the function before calling search_paths again
                        bb = B0, current_path = [B0], current_call_stack = [(None, Main)], current_subroutine_executed = [[B0]]

                    # B0 calls Sa. Calling a subroutine adds a item to current_call_stack and current_subroutine_executed.
                    Start
                        bb = B3, current_path = [B0], current_call_stack = [(None, Main), (B0, Sa)], current_subroutine_executed = [[B0], []]
                    End
                        bb = B3, current_path = [B0, B3], current_call_stack = [(None, Main), (B0, Sa)], current_subroutine_executed = [[B0], [B3]]

                    # B3 calls Sb. Calling a subroutine adds a item to current_call_stack and current_subroutine_executed.
                    Start
                        bb = B5, current_path = [B0, B3], current_call_stack = [(None, Main), (B0, Sa), (B3, Sb)], current_subroutine_executed = [[B0], [B3], []]
                    End
                        bb = B5, current_path = [B0, B3, B5], current_call_stack = [(None, Main), (B0, Sa), (B3, Sb)], current_subroutine_executed = [[B0], [B3], [B5]]

                    # B5 returns the execution from Sb. Returning from a subroutine removes a item from current_call_stack, current_subroutine_executed.
                    Start
                        bb = B4, current_path = [B0, B3, B5], current_call_stack = [(None, Main), (B0, Sa)], current_subroutine_executed = [[B0], [B3]]
                    End
                        bb = B4, current_path = [B0, B3, B5, B4], current_call_stack = [(None, Main), (B0, Sa)], current_subroutine_executed = [[B0], [B3, B4]]

                    Start
                        bb = B1, current_path = [B0, B3, B5, B4], current_call_stack = [(None, Main)], current_subroutine_executed = [[B0]]
                    End
                        bb = B1, current_path = [B0, B3, B5, B4, B1], current_call_stack = [(None, Main)], current_subroutine_executed = [[B0, B1]]

                    ....
        """
        # check for loops
        if bb in current_subroutine_executed[-1]:
            logger_detectors.debug(
                f"Loop Path: current_full_path = {current_path}\n, current_subroutine_executed = {current_subroutine_executed[-1]}, current_block: {repr(bb)}"
            )
            return

        if validated_in_block(bb, function, checks_field):
            logger_detectors.debug(
                f"Validated Path: current_full_path = {current_path}\n, current_block: {repr(bb)}"
            )
            return

        current_path = current_path + [bb]

        # leaf block
        if leaf_block_global(bb):
            logger_detectors.debug(f"Vulnerable Path: current_full_path = {current_path}")
            if satisfies_report_condition(current_path):
                # only report if it satisfies the report condition
                logger_detectors.debug(
                    f"Vulnerable Path Satisfied Report Condition: current_full_path = {current_path} "
                )
                paths_without_check.append(current_path)
            return

        # do we need to make a copy of lists in [:-1]???
        current_subroutine_executed = current_subroutine_executed[:-1] + [
            current_subroutine_executed[-1] + [bb]
        ]

        if bb.is_callsub_block:
            called_subroutine = bb.called_subroutine
            # check for recursion
            already_called_subroutines = [frame[1] for frame in current_call_stack]
            if called_subroutine in already_called_subroutines:
                # recursion
                return
            current_call_stack = current_call_stack + [(bb, called_subroutine)]
            current_subroutine_executed = current_subroutine_executed + [[]]

        if bb.is_retsub_block:
            # if this block is retsub then it returns execution to the return
            # point of callsub block. return point is the next instruction after the callsub
            # instruction.
            (callsub_block, _) = current_call_stack[-1]
            assert callsub_block is not None
            return_point = callsub_block.sub_return_point
            if return_point is not None:
                search_paths(
                    return_point,
                    current_path,
                    paths_without_check,
                    current_call_stack[:-1],  # returning from a subroutine
                    current_subroutine_executed[:-1],
                )
        else:
            for next_bb in next_blocks_global(function, bb):
                search_paths(
                    next_bb,
                    current_path,
                    paths_without_check,
                    current_call_stack,
                    current_subroutine_executed,
                )

    entry_block = function.entry
    paths_without_check: List[List["BasicBlock"]] = []
    search_paths(entry_block, [], paths_without_check, [(None, function.main)], [[]])

    return paths_without_check


def detect_missing_tx_field_validations_group(
    tealer: "Tealer",
    checks_field: Callable[["BlockTransactionContext"], bool],
    satisfies_report_condition: Callable[[List["BasicBlock"]], bool] = lambda _x: True,
) -> List[Tuple["Teal", List[List["BasicBlock"]]]]:
    """Given the tealer object, return vulnerable execution paths.
    The current implementation ignores the transaction structure. It iterates over the

    Args:
        tealer: the tealer object
        checks_field:  A function which given a block context, should return True if the target field
            cannot have the vulnerable value or else False.
        satisfies_report_condition: Given a path, should return True if the "path" satifies the vulnerable condition.

    Returns:
        Returns a list of vulnerable execution paths: none of the blocks in the path check the fields.
    """
    output: List[Tuple["Teal", List[List["BasicBlock"]]]] = []
    for group_txn in tealer.groups:
        # Current implementation only works for single contract invocations
        # assert len(group_txn.transactions) == 1
        for txn in group_txn.transactions:
            if txn.has_logic_sig and txn.logic_sig is not None:
                vulnerable_paths = detect_missing_tx_field_validations(
                    txn.logic_sig, checks_field, satisfies_report_condition
                )
                # Using old output class for now.
                # TODO: create new output class to represent vulnerable paths in a group transaction context.
                contract = txn.logic_sig.contract
                output.append((contract, vulnerable_paths))
            if txn.application is not None:
                vulnerable_paths = detect_missing_tx_field_validations(
                    txn.application, checks_field, satisfies_report_condition
                )
                contract = txn.application.contract
                output.append((contract, vulnerable_paths))
    return output


def contract_checks_its_field(
    function: "Function",
    checks_field: Callable[["BlockTransactionContext"], bool],
    absolute_index: Optional[int],
) -> bool:
    """
    Return True if the contract checks its transaction field using `txn {field}` or `gtxn(s) {fields}` by accessing it using the index.

    Args:
        function: The function being checked
        checks_field: A function which returns True if function validates the vulnerable transaction "field".
        absolute_index: The absolute index of the transaction given by the user in the config.

    Returns:
        Returns True if the contracts checks its own transaction field.
    """
    leaf_blocks = [block for block in function.blocks if leaf_block_global(block)]
    for block in leaf_blocks:
        # checking the leaf blocks is enough to confirm the vulnerability
        if not validated_in_block(block, function, checks_field, absolute_index):
            # is vulnerable
            return False
    # contract checks its fields
    return True


def contract_checks_txn_at_absolute_index(
    function: "Function",
    checks_field: Callable[["BlockTransactionContext"], bool],
    absolute_index: int,
) -> bool:
    """
    Returns True if the Function checks the field of the transaction at `absolute_index` else False

    Args:
        function: The function
        checks_field: A function which returns True if the context information has the vulnerable value.
        absolute_index: The absolute index of the other transaction

    Returns:
        Returns True if the function checks the field of Gtxn[absolute_index] else False
    """
    leaf_blocks = [block for block in function.blocks if leaf_block_global(block)]
    for block in leaf_blocks:
        if not checks_field(function.transaction_context(block).absolute_context(absolute_index)):
            # There is an execution path from entry to this leaf block in which the transaction field can have the
            # vulnerable value.
            return False
    # The transaction field of Gtxn[absolute_index] is checked in all paths.
    return True


def contract_checks_using_relative_index(
    function: "Function",
    checks_field: Callable[["BlockTransactionContext"], bool],
    offset: int,
) -> bool:
    """
    Returns True if the Function checks the field of the transaction at `offset` from the function else False

    Args:
        function: The function
        checks_field: A function which returns True if the context information has the vulnerable value.
        offset: The offset of the other transaction from this function

    Returns:
        Returns True if the function checks the field of Gtxn[Txn.group_index() + offset] else False
    """
    leaf_blocks = [block for block in function.blocks if leaf_block_global(block)]
    for block in leaf_blocks:
        if not checks_field(function.transaction_context(block).relative_context(offset)):
            # There is an execution path from entry to this leaf block in which the transaction field can have the
            # vulnerable value.
            return False
    # The transaction field of Gtxn[absolute_index] is checked in all paths.
    return True


# def get_leaf_blocks(function: "Function") -> List["BasicBlock"]:
#     for
# change to a better name
def detect_missing_tx_field_validations_group_complete(  # pylint: disable=too-many-branches
    tealer: "Tealer",
    detector: "AbstractDetector",
    checks_field: Callable[["BlockTransactionContext"], bool],
    vulnerable_transaction_types: Optional[List[TransactionType]] = None,
) -> List[GroupTransactionOutput]:
    """
    `tealer.analyses.dataflow` calculates possible values a transaction field can have to successfully complete execution.
    Information is computed at block level. Each block's context(`BlockTransactionContext`)
    will have information on possible values for a transaction field such that
        - Execution reaches to the first instruction of block `B`
        - Execution reaches exit instruction of block `B` without failing.

    The goal of current detectors is to find if a given transaction field(s) can have a certain vulnerable value(s).
    e.g: is_updatable detector checks if the transaction can be UpdateApplication type transaction. Here, the vulnerable value for
    transaction type is "UpdateApplication". if "UpdateApplication" is a possible value then the contract is considered vulnerable and is reported.

    In order to find whether the given contract is vulnerable, it is sufficient to check the leaf blocks: whether transaction field can
    have target value and execution reaches end of a leaf block. Note, this is not always true for AnyoneCanUpdate and AnyoneCanDelete dectectors. These
    detectors depend on two fields that are tracked by independent keys. However, the current implementation is sufficient for the common patterns.
    - Only LogicSigs are vulnerable to Stateless detectors.
    - If a transaction does not have LogicSig then it is considered to be not vulnerable by Stateless detectors.

    Args:
        tealer: the tealer object
        detector: The detector
        checks_field:  A function which given a block context, should return True if the target field
            cannot have the vulnerable value or else False.
        vulnerable_transaction_types: The contracts are vulnerable when it is executed in only some of the transaction types. This list is that list of
            transaction types. If this list is given, the transactions whose transaction type is not in the list will be skipped and are considered not
            vulnerable to the detector.

    Returns:
        Found vulnerable operations/groups
    """
    output = []
    for group_txn in tealer.groups:
        is_vulnerable = False
        vulnerable_transactions = {}
        for txn in group_txn.transactions:
            if detector.TYPE == DetectorType.STATELESS and not txn.has_logic_sig:
                continue

            if detector.TYPE == DetectorType.STATEFULL and txn.application is None:
                # check for the application transaction type???
                continue

            if (
                vulnerable_transaction_types is not None
                and txn.type not in vulnerable_transaction_types
            ):
                continue
            # print(txn.logic_sig.blocks)
            # for block in txn.logic_sig.blocks:
            #     print(block)
            #     print(leaf_block_global(block))
            #     print(txn.logic_sig.transaction_context(block).rekeyto)
            # the contract has logic sig
            # check if the contracts executed in the txn check the field.
            # logic sig contract is given
            # print("A")
            if txn.logic_sig is not None and contract_checks_its_field(
                txn.logic_sig, checks_field, txn.absoulte_index
            ):
                # the logic sig checks its field, not vulnerable
                continue
            # print("B")
            if txn.application is not None and contract_checks_its_field(
                txn.application, checks_field, txn.absoulte_index
            ):
                # the application checks the field. The transaction is not vulnerable
                # continue to next transaction
                continue
            # check if any other transaction checks this transaction's field using absolute index
            if txn.absoulte_index is not None:
                checked = False
                for other_txn in group_txn.transactions:
                    if other_txn.logic_sig is not None and contract_checks_txn_at_absolute_index(
                        other_txn.logic_sig, checks_field, txn.absoulte_index
                    ):
                        # The logic sig of some transaction in the group checks this transaction index using the absolute index
                        checked = True
                        break
                    if other_txn.application is not None and contract_checks_txn_at_absolute_index(
                        other_txn.application, checks_field, txn.absoulte_index
                    ):
                        checked = True
                        break

                if checked:
                    continue

            # check if any other transaction in the group checks this transaction's field using relative index
            # group_txn.group_relative_indexes[t1][t2] = -1 => t1.group_index() == t2.group_index() - 1
            # txn_relative_indexes =  group_txn.group_relative_indexes[txn]
            checked = False
            for (other_txn, offset) in group_txn.group_relative_indexes[txn].items():
                # offset of this txn from other_txn. other_txn would have used this offset to access the field and validate it.
                if other_txn.logic_sig is not None and contract_checks_using_relative_index(
                    other_txn.logic_sig, checks_field, offset
                ):
                    # The logic sig of some other transaction in the group checks this transaction usiing the relative index.
                    checked = True
                    break

                if other_txn.application is not None and contract_checks_using_relative_index(
                    other_txn.application, checks_field, offset
                ):
                    checked = True
                    break

            if checked:
                continue

            # the contract is vulnerable. mark the operation/group as vulnerable
            is_vulnerable = True

            if detector.TYPE == DetectorType.STATELESS:
                if txn.logic_sig is not None:
                    vulnerable_transactions[txn] = [txn.logic_sig]
                else:
                    vulnerable_transactions[txn] = []
            elif detector.TYPE == DetectorType.STATEFULL:
                if txn.application is not None:
                    vulnerable_transactions[txn] = [txn.application]
                else:
                    vulnerable_transactions[txn] = []

        if is_vulnerable:
            output.append(GroupTransactionOutput(detector, group_txn, vulnerable_transactions))
    return output



"""Util functions to output dot files and detector results.

This modules contains functions, classes which are used to store
and display different types of output formats used by tealer detectors
and printers.

Functions:
    full_cfg_to_dot(teal: "Teal", config: Optional[CFGDotConfig]=None, filename: Optional[Path]=None) -> None:
        Exports dot representation of CFG represented by :bbs: in
        dot format to given filename.

Classes:
    ExecutionPaths: Class to represent results of a detector, stores
        execution paths detected by the detector.

Types:
    SupportedOutput: Union of types used for representing detector results.
        For now, it is an alias for ExecutionPaths.

"""
import abc
import html
import re
import os
from pathlib import Path
from typing import List, TYPE_CHECKING, Dict, Callable, Optional
from dataclasses import dataclass, field

from tealer.teal.instructions.instructions import BZ, BNZ, Callsub, Retsub

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.subroutine import Subroutine
    from tealer.teal.functions import Function
    from tealer.teal.teal import Teal
    from tealer.teal.instructions.instructions import Instruction
    from tealer.detectors.abstract_detector import AbstractDetector
    from tealer.execution_context.transactions import GroupTransaction, Transaction


ROOT_OUTPUT_DIRECTORY = Path(os.getenv("TEALER_ROOT_OUTPUT_DIR", "tealer-export"))


@dataclass
class CFGDotConfig:  # pylint: disable=too-many-instance-attributes
    ins_additional_comments: Callable[["Instruction"], List[str]] = lambda _x: []
    # include additional tealer comments at the top of the block.
    bb_additional_comments: Callable[["BasicBlock"], List[str]] = lambda _x: []
    # don't include edge bi -> bj in dot output?
    ignore_edge: Callable[["BasicBlock", "BasicBlock"], bool] = lambda _bi, _bj: False
    # Apply colors to different types of edges?
    color_edges: bool = True
    jump_branch_color: str = "#36d899"  # "green"
    default_branch_color: str = "#e0182b"  # "red"
    callsub_edge_color: str = "#ff8c00"  # "orange"
    remaining_edges_color: str = "BLACK"
    comments_cell_border_size: int = 2  # size of basic block comments cell
    bb_border_color: Callable[["BasicBlock"], str] = lambda _x: "BLACK"
    custom_background_color: Dict["Instruction", str] = field(default_factory=dict)


def _instruction_to_dot(ins: "Instruction", config: CFGDotConfig) -> str:
    """Return dot representation of Teal instruction.

    string representation of the instruction is represented as
    a table cell(row) in dot.

    Args:
        ins: teal instruction to represent in dot format.
        config: configuration for dot output

    Returns:
        string containing the dot representation of the given
        instruction.

    ---------------
    // tealer comments
    //
    // source code comments
    // ...
    // instruction
    ---------------
    """

    # ins.source_code stores the indentation and whitespace. strip them
    ins_str = html.escape(ins.source_code.strip(), quote=True)  # original teal code
    # make callsub and retsub bold and italic
    if isinstance(ins, (Callsub, Retsub)):
        ins_str = f"<B><I>{ins_str}</I></B>"

    # format:
    # <B>
    #   // tealer_comment_1 <BR/>
    #   // tealer_comment_2 <BR/>
    # </B>
    # <BR/>
    tealer_comments = ""
    if ins.tealer_comments + config.ins_additional_comments(ins):
        sanitized_comments = [
            html.escape(comment.strip(), quote=True)
            for comment in ins.tealer_comments + config.ins_additional_comments(ins)
        ]
        tealer_comments = "<BR/>".join(f"// {comment}" for comment in sanitized_comments)
        tealer_comments = f"<B>{tealer_comments}</B><BR/>"  # make them bold

    # format:
    #   // source_comment_1 <BR/>
    #   // source_comment_2 <BR/>
    source_code_comments = ""
    if ins.comments_before_ins:
        sanitized_comments = [
            html.escape(comment.strip(), quote=True) for comment in ins.comments_before_ins
        ]
        source_code_comments = "<BR/>".join(f"{comment}" for comment in sanitized_comments)
        source_code_comments += "<BR/>"

    color = config.custom_background_color.get(ins, "BLACK")

    cell_i = (
        "<TR>"
        f'<TD ALIGN="LEFT" BALIGN="LEFT" COLOR="{color}">'
        f"{tealer_comments}"
        f"{source_code_comments}"
        f"{ins.line}. {ins_str}"
        "</TD>"
        "</TR>\n"
    )
    return cell_i


def _bb_to_dot(bb: "BasicBlock", config: CFGDotConfig) -> str:
    """Return dot representation of basic block.

    Basic Blocks are represented in the form of a tabel in dot.
    Each instruction in the basic block is represented as a row.

    Args:
        bb: basic block to represent in dot format.
        config: configuration for dot output

    Returns:
        string containing the dot representation of the given
        basic block.
    """

    # format: `{source_node}:s -> {dest_node}{dest_port}:n [color=""];`
    def graph_edge_str(src_bb: "BasicBlock", dest_bb: "BasicBlock", edge_color: str) -> str:
        if config.ignore_edge(src_bb, dest_bb):
            return ""
        return f'{src_bb.idx}:s -> {dest_bb.idx}:{dest_bb.entry_instr.line}:n [color="{edge_color}"];\n'

    # format:
    #     <TR> tealer_comments and additional_comments </TR>
    #     <TR> instruction_1                           </TR>
    #     <TR> instruction_2                           </TR>
    #     ...
    table_rows: List[str] = []

    santized_comments = [
        html.escape(comment.strip(), quote=True)
        for comment in bb.tealer_comments + config.bb_additional_comments(bb)
    ]
    comments_cell_str = (
        "<TR>"
        f'<TD COLOR="BLACK" ALIGN="LEFT" BALIGN="LEFT" PORT="{bb.entry_instr.line}" BORDER="{config.comments_cell_border_size}">'
        "<B>"
        f'{"<BR/>".join(f"// {comment}" for comment in santized_comments)}'
        "</B>"
        "</TD>"
        "</TR>\n"
    )

    table_rows.append(comments_cell_str)

    for ins in bb.instructions:
        table_rows.append(_instruction_to_dot(ins, config))

    graph_edges: List[str] = []
    if config.color_edges and isinstance(bb.exit_instr, (BZ, BNZ)):
        if isinstance(bb.exit_instr, (BZ, BNZ)):
            # color graph edges if exit instruction is BZ or BNZ.
            if len(bb.next) == 1:
                # happens when bz/bnz is the last instruction in the contract and there is no default branch
                default_branch = None
                jump_branch = bb.next[0]
            else:
                default_branch = bb.next[0]
                jump_branch = bb.next[1]

            if default_branch is not None:
                graph_edges.append(graph_edge_str(bb, default_branch, config.default_branch_color))
            graph_edges.append(graph_edge_str(bb, jump_branch, config.jump_branch_color))
    else:
        for next_bb in bb.next:
            graph_edges.append(graph_edge_str(bb, next_bb, config.remaining_edges_color))

    table_str = (
        f'<<TABLE ALIGN="LEFT" COLOR="{config.bb_border_color(bb)}">\n'
        f'{"".join(table_rows)}'
        "</TABLE>> labelloc=top shape=plain\n"
    )

    return f'{bb.idx}[label={table_str}] {"".join(graph_edges)}'


def subroutine_to_dot(subroutine: "Subroutine", config: Optional[CFGDotConfig] = None) -> str:
    if config is None:
        config = CFGDotConfig()
    # ignore edges from callsub to the return point block because we place an empty node to represent
    # the called subroutine in between the callsub block and the return point.
    # replacing the ignore_edge function directly should be ok (?) for now
    config.ignore_edge = lambda bi, _: isinstance(bi.exit_instr, (Callsub))

    def empty_subroutine_box(callsub_block: "BasicBlock") -> str:
        # Include a empty box between callsub and it's return point. Empty box represents
        # the subroutine called by `callsub`.
        called_subroutine = callsub_block.called_subroutine
        return_point_block = callsub_block.sub_return_point
        if return_point_block is None:
            # happens when callsub is the last instruction in the CFG and subroutine always exits the program.
            return_point_block_idx = "none"
        else:
            return_point_block_idx = str(return_point_block.idx)
        content = (
            f'"Subroutine {called_subroutine.name}"'  # TODO: add other information in the future
        )
        node_name = f"x{callsub_block.idx}_{return_point_block_idx}"
        edge1 = f"{callsub_block.idx}:s -> {node_name}:n;\n"  # callsub to empty box
        edge2 = ""
        if return_point_block is not None:
            # edge from empty box to return point block
            edge2 = f"{node_name}:s -> {return_point_block.idx}:{return_point_block.entry_instr.line}:n;\n"

        return f"{node_name}[label={content},style=dashed,shape=box,fontname=bold] {edge1}{edge2}"

    nodes_dot: List[str] = []
    for bi in subroutine.blocks:
        nodes_dot.append(_bb_to_dot(bi, config))
        if bi.is_callsub_block:
            # add empty box to represent the graph of called subroutine
            # add edge from callsub to that box and box to callsub return point.
            # ignoring recursion here. adds empty box even if callsub calls the same subroutine.
            nodes_dot.append(empty_subroutine_box(bi))

    nodes_str = "\n".join(nodes_dot)

    dot_output = "digraph g{\n ranksep = 1 \n overlap = scale \n" f"{nodes_str}\n" "}"  # .......

    return dot_output


def all_subroutines_to_dot(
    teal: "Teal",
    dest: Path,
    config: Optional[CFGDotConfig] = None,
    filename_prefix: str = "",
) -> None:
    """Export CFG of each subroutine to a dot file.

    Args:
        teal: contract instance
        dest: destination directory to save the dot files.
        config: CFGDotConfig to use for generating the dot files. if None, function uses
            default settings.
        filename_prefix: prefix to add before each filename to distinguish files generated
            by multiple calls to this function.
    """
    if filename_prefix:  # not empty string
        filename_prefix = f"{filename_prefix}_"
    main_entry_sub_filename = f"{filename_prefix}contract_shortened_cfg.dot"

    with open(dest / Path(main_entry_sub_filename), "w", encoding="utf-8") as f:
        f.write(subroutine_to_dot(teal.main, config))
        print(f"Exported contract's shortened cfg to: {dest / Path(main_entry_sub_filename)}")

    for sub_name, subroutine in teal.subroutines.items():
        filename = f"{filename_prefix}subroutine_{sub_name}_cfg.dot"
        with open(dest / Path(filename), "w", encoding="utf-8") as f:
            f.write(subroutine_to_dot(subroutine, config))
            print(f'Exported cfg of "{sub_name}" subroutine to: {dest / Path(filename)}')


def full_cfg_to_dot(  # pylint: disable=too-many-locals
    teal: "Teal", config: Optional[CFGDotConfig] = None, filename: Optional[Path] = None
) -> Optional[str]:
    """Export control flow graph to a dot file.

    The control flow graph is represented as a digraph in dot.
    basic blocks are represented as a table with it's instructions
    as rows.

    Args:
        teal: The contract.
        config: optional configuration for dot output.
        filename: name of the file to save the dot representation
            of control flow graph in.

    Returns:
        Returns dot representation of the CFG if filename is not given. If filename is given,
        writes the dot representation to the file.
    """

    bbs = teal.bbs
    subroutine_block_idx = set(bb.idx for bb in teal.bbs if bb in teal.main.blocks)
    # responsible for "box" around each subroutine.
    subroutine_clusters: List[str] = []
    for i, (subroutine_name, subroutine) in enumerate(teal.subroutines.items()):
        subroutine_bbs = subroutine.blocks
        cluster_name = i
        cluster_nodes = " ".join(str(bb.idx) for bb in subroutine_bbs)
        # TODO: Add number of args and return values to label in the future.
        subgraph_dot = f"""
            subgraph cluster_{cluster_name} {{
                label = "Subroutine {subroutine_name}";
                graph[style=dashed];
                {cluster_nodes};
            }}
        """
        subroutine_clusters.append(subgraph_dot)

    # format: `{source_node}:s -> {dest_node}{dest_port}:n [color=""];`
    def graph_edge_str(src_bb: "BasicBlock", dest_bb: "BasicBlock", edge_color: str) -> str:
        return f'{src_bb.idx}:s -> {dest_bb.idx}:{dest_bb.entry_instr.line}:n [color="{edge_color}"];\n'

    # default config
    if not config:
        config = CFGDotConfig()
        # border color for subroutine blocks
        subroutine_blocks_border_color = "#000066"
        config.bb_border_color = (
            lambda bb: "BLACK"
            if bb.idx not in subroutine_block_idx
            else subroutine_blocks_border_color
        )

    # ignore callsub block to return point edges. Add edges from callsub blocks to subroutine entry and retsubs to return points
    config.ignore_edge = lambda bi, _: isinstance(bi.exit_instr, (Callsub))

    bb_nodes_dot: List[str] = []
    for bb in bbs:
        bb_nodes_dot.append(_bb_to_dot(bb, config))
        if bb.is_callsub_block:
            # add edge from callsub block to subroutine entry
            bb_nodes_dot.append(
                graph_edge_str(bb, bb.called_subroutine.entry, config.callsub_edge_color)
            )
            # add edge from retsub blocks to return point
            return_point_block = bb.sub_return_point
            if return_point_block is None:
                continue
            for src_bb in bb.called_subroutine.retsub_blocks:
                bb_nodes_dot.append(
                    graph_edge_str(src_bb, return_point_block, config.remaining_edges_color)
                )

    subroutine_clusters_str = "\n".join(subroutine_clusters)
    bb_nodes_str = "\n".join(bb_nodes_dot)

    dot_output = (
        "digraph g{\n ranksep = 1 \n overlap = scale \n"
        f"{subroutine_clusters_str}\n"
        f"{bb_nodes_str}\n"
        "}"
    )

    if filename is None:
        return dot_output

    with open(filename, "w", encoding="utf-8") as f:
        f.write(dot_output)
    return None


def detector_terminal_description(detector: "AbstractDetector") -> str:
    """Return description for the detector that is printed to terminal before listing vulnerable paths.

    Args:
        detector: A detector object.

    Returns:
        Returns description for the :detector: that can be printed on the terminal or used as general description
        for the detector.
    """
    return (
        f'\nCheck: "{detector.NAME}", Impact: {detector.IMPACT}, Confidence: {detector.CONFIDENCE}\n'
        f"Description: {detector.DESCRIPTION}\n\n"
        f"Wiki: {detector.WIKI_URL}\n"
    )


def detector_ouptut_dir(destination: Path, detector: "AbstractDetector") -> Path:
    return destination / Path(detector.NAME)


class Output(abc.ABC):
    @property
    @abc.abstractmethod
    def detector(self) -> "AbstractDetector":
        pass

    @abc.abstractmethod
    def filter_paths(self, filter_regex: str) -> None:
        pass

    @abc.abstractmethod
    def to_json(self) -> Dict:
        pass

    @abc.abstractmethod
    def generate_output(self, dest: Path) -> bool:
        """
        Generate the output


        Args:
            dest: The files will be saved in the given :dest: destination directory.

        Returns:
            Returns true if something was generated - False if there is nothing to be written
        """

        return False  # this statement is needed for darglint


class InstructionsOutput(Output):
    def __init__(
        self, teal: "Teal", detector: "AbstractDetector", instructions: List[List["Instruction"]]
    ):
        self._teal = teal
        self._detector = detector
        self.instructions: List[List["Instruction"]] = instructions

    @property
    def detector(self) -> "AbstractDetector":
        return self._detector

    def filter_paths(self, filter_regex: str) -> None:
        pass

    def to_json(self) -> Dict:
        result = {
            "type": "InstructionsOutput",
            "count": len(self.instructions),
            "description": detector_terminal_description(self.detector),
            "check": self.detector.NAME,
            "impact": str(self.detector.IMPACT),
            "confidence": str(self.detector.CONFIDENCE),
            "help": self.detector.WIKI_RECOMMENDATION.strip(),
            "paths": [str(ins) for ins in self.instructions],
        }
        return result

    def generate_output(self, dest: Path) -> bool:
        """
        Generate the output


        Args:
            dest: Not use (no files are generated for InstructionsOutput)

        Returns:
            Returns true if something was generated - False if there is nothing to be written
        """

        if not self.instructions:
            return False

        print(detector_terminal_description(self.detector))
        print("\tFollowing are the unoptimized instructions found:")

        for ins in self.instructions:
            print(ins)

        return True


class ExecutionPaths(Output):
    """Detector output class to represent vulnerable execution paths."""

    def __init__(self, teal: "Teal", detector: "AbstractDetector", paths: List[List["BasicBlock"]]):
        self._teal = teal
        self._detector = detector
        self.paths: List[List["BasicBlock"]] = paths

    @property
    def detector(self) -> "AbstractDetector":
        return self._detector

    def _filename(self, path_index: int) -> Path:
        return Path(f"{self.detector.NAME}-{path_index}.dot")

    @staticmethod
    def _short_notation(path_bbs: List["BasicBlock"]) -> str:
        """Return short notation representation of path

        Args:
            path_bbs: List of basic blocks in the path from the contract's entry.

        Returns:
            Returns short notation representation of the path.
            if path has [B0, B2, B3, B5] then short notation is "0 -> 2 -> 3 -> 5"
        """
        # TODO: Change notation from "0 -> 2 -> 3 -> 5" to "B0 -> B2 -> B3 -> B5"
        return " -> ".join(map(str, [bb.idx for bb in path_bbs]))

    def filter_paths(self, filter_regex: str) -> None:
        if filter_regex == "":
            return
        filtered_paths: List[List["BasicBlock"]] = []
        for path in self.paths:
            if re.search(filter_regex, self._short_notation(path)) is None:
                # short notation does not contain string matching the regex
                filtered_paths.append(path)
        self.paths = filtered_paths
        return

    def generate_output(self, dest: Path) -> bool:
        """Export execution paths to dot files.

        The execution paths are highlighted in the dot representation
        of CFG. Each execution path is indexed based on the order they
        are added to the result and index will be used in the filename.

        Args:
            dest: The dot files will be saved in the given :dest: destination
                directory.

        Returns:
            Returns true if something was written - False if there is nothing to be written
        """

        if not self.paths:
            return False

        print(detector_terminal_description(self.detector))

        # cfg_to_dot config
        config = CFGDotConfig()
        config.color_edges = False
        print("\tFollowing are the vulnerable paths found:")

        dest = detector_ouptut_dir(dest, self.detector)
        # create output directory if not present
        os.makedirs(dest, exist_ok=True)

        for idx, path in enumerate(self.paths, start=1):
            short = self._short_notation(path)
            print(f"\n\t\t path: {short}")

            filename = dest / self._filename(idx)
            print(f"\t\t check file: {filename}")

            config.bb_border_color = (
                lambda bb: "BLACK"
                if bb not in path  # pylint: disable=cell-var-from-loop
                else "RED"
            )
            full_cfg_to_dot(self._teal, config, filename)
        print("-" * 100)

        return True

    def to_json(self) -> Dict:
        """Return json representation of detector result.

        The execution paths are represented as a list of basic blocks,
        which themselves are list of string representation of it's
        instructions.

        Returns:
            JSON encodable dictionary representing the detector result.
        """

        result = {
            "type": "ExecutionPaths",
            "count": len(self.paths),
            "description": detector_terminal_description(self.detector),
            "check": self.detector.NAME,
            "impact": str(self.detector.IMPACT),
            "confidence": str(self.detector.CONFIDENCE),
            "help": self.detector.WIKI_RECOMMENDATION.strip(),
        }
        paths = []
        for path in self.paths:
            short = " -> ".join(map(str, [bb.idx for bb in path]))
            blocks = []
            for bb in path:
                block = []
                for ins in bb.instructions:
                    block.append(f"{ins.line}: {ins}")
                blocks.append(block)

            paths.append({"short": short, "blocks": blocks})

        result["paths"] = paths
        return result


class GroupTransactionOutput(Output):
    def __init__(
        self,
        detector: "AbstractDetector",
        group: "GroupTransaction",
        transactions: Dict["Transaction", List["Function"]],
    ) -> None:
        self._detector = detector
        self._group = group
        self._transactions = transactions

    @property
    def detector(self) -> "AbstractDetector":
        return self._detector

    @property
    def group_transaction(self) -> "GroupTransaction":
        return self._group

    @property
    def transactions(self) -> Dict["Transaction", List["Function"]]:
        return self._transactions

    def filter_paths(self, filter_regex: str) -> None:
        pass

    def to_json(self) -> Dict:
        transactions = {}
        for txn in self._transactions:
            contracts = []
            for function in self._transactions[txn]:
                contracts.append(
                    {
                        "contract": function.contract.contract_name,
                        "function": function.function_name,
                    }
                )
            transactions[txn.transacton_id] = contracts

        result = {
            "type": "GroupTransactionOutput",
            "description": detector_terminal_description(self.detector),
            "check": self.detector.NAME,
            "impact": str(self.detector.IMPACT),
            "confidence": str(self.detector.CONFIDENCE),
            "help": self.detector.WIKI_RECOMMENDATION.strip(),
            "operation": self._group.operation_name,
            "transactions": transactions,
        }
        return result

    def generate_output(self, dest: Path) -> bool:
        """
        Generate the output

        Args:
            dest: Not used, no files are generated for GroupTransactionOutput

        Returns:
            Returns true if something was generated - False if there is nothing to be written.
        """
        print(detector_terminal_description(self.detector))
        print(
            f"\tFollowing transactions of the operation {self._group.operation_name} are vulnerable:\n"
        )

        for txn in self._transactions:
            print(f"\tTransaction {txn.transacton_id}")
            if self._transactions[txn]:
                # contract was given by the user.
                for function in self._transactions[txn]:
                    print(f"\t\tContract: {function.contract.contract_name}")
                    print(f"\t\tFunction: {function.function_name}")
                    print("\n")

        return True


ListOutput = List[Output]



"""Defines classes to represent app_params_get fields.

``app_params_get`` instruction is used to access parameter fields
of an application in the contract.

Each field that can be accessed using app_params_get is represented
by a class in tealer. All the classes representing the fields must
inherit from AppParamsField class.

"""

# pylint: disable=too-few-public-methods
class AppParamsField:
    """Base class to represent App fields."""

    def __init__(self) -> None:
        self._version: int = 5

    @property
    def version(self) -> int:
        """Teal version this field is introduced in and supported from.

        Returns:
            Teal version the field is introduced in and supported from.
        """
        return self._version

    def __str__(self) -> str:
        return self.__class__.__qualname__


class AppApprovalProgram(AppParamsField):
    """Bytecode of approval program of the App."""


class AppClearStateProgram(AppParamsField):
    """Bytecode of clear state program of the App."""


class AppGlobalNumUint(AppParamsField):
    """Number of uint64 values allowed in Global State."""


class AppGlobalNumByteSlice(AppParamsField):
    """Number of byte array values allowed in Global State."""


class AppLocalNumUint(AppParamsField):
    """Number of uint64 values allowed in Local State."""


class AppLocalNumByteSlice(AppParamsField):
    """Number of byte array values allowed in Local State."""


class AppExtraProgramPages(AppParamsField):
    """Number of extra program pages of code space."""


class AppCreator(AppParamsField):
    """Address of Creator(deployer) of the application."""


class AppAddress(AppParamsField):
    """Address for which this application has authority."""



from typing import TYPE_CHECKING, Tuple, Optional
from dataclasses import dataclass

from tealer.teal.instructions.instructions import (
    Txn,  # self, normal field
    Gtxn,  # txn index is part of the instruction, normal field
    # Txna,   # self, array field, array index in the instruction
    # Gtxna,  # txn index is part of the instruction, array field, array index in the instruction
    Gtxns,  # txn index is taken from stack, normal field
    # Gtxnsa, # txn index is taken from stack, normal field, array index in the instruction
    # Txnas,  # self, array field, array index from stack
    # Gtxnas, # txn index is part of the instruction, array field, array index from stack
    # Gtxnsas, # txn index from stack, array field, array index from stack.
    Sub,
    Add,
)
from tealer.utils.comparable_enum import ComparableEnum
from tealer.analyses.utils.stack_ast_builder import KnownStackValue, UnknownStackValue
from tealer.teal.instructions.transaction_field import GroupIndex
from tealer.utils.analyses import is_int_push_ins

if TYPE_CHECKING:
    from tealer.teal.instructions.transaction_field import TransactionField


class IndexType(ComparableEnum):
    Self = 0
    Absolute = 1
    Relative = 2

    Unknown = 777


@dataclass
class TransactionIndex:
    index_type: IndexType
    # value should be ignored for Self and Unknown index types.
    value: int


def get_index_and_field(
    value: KnownStackValue,
) -> Tuple[bool, Optional["TransactionIndex"], Optional["TransactionField"]]:
    """Given a stack value, check if the value pushed is a transaction field. if so, return transaction index and the field.

    Args:
        value: The stack value.

    Returns:
        bool: True if the pushed value is a transaction field else False
        TransactionIndex: Index of the transaction
        TransactionField: The transaction field that will be pushed.
    """
    # TODO: probably could cache
    # our analysis does not concern array fields, so ignoring those instructions for now.
    if not isinstance(value.instruction, (Txn, Gtxn, Gtxns)):
        return False, None, None

    field = value.instruction.field
    if isinstance(value.instruction, Txn):
        return True, TransactionIndex(IndexType.Self, 0), field

    if isinstance(value.instruction, Gtxn):
        return True, TransactionIndex(IndexType.Absolute, value.instruction.idx), field

    # Gtxns instruction
    index_stack_value = value.args[0]
    # happens when the Gtxns instruction is the first instruction of the block.
    if isinstance(index_stack_value, UnknownStackValue):
        return True, TransactionIndex(IndexType.Unknown, 0), field

    index = _get_index(index_stack_value)
    return True, index, field


def _get_index(index_stack_value: KnownStackValue) -> "TransactionIndex":
    # index patterns
    # Txn.group_index() -> Self
    # Int(n) -> Absolute index
    # Txn.group_index() +/- Int(n) -> Relative / Unknown

    # TODO: if there was some kind of analysis to calculate index implemented, call that function to fetch the index directly.
    # There is no need for such analysis. below patterns cover most of the contracts.
    # Txn.group_index() -> txn GroupIndex
    if isinstance(index_stack_value.instruction, Txn) and isinstance(
        index_stack_value.instruction.field, GroupIndex
    ):
        # Self
        return TransactionIndex(IndexType.Self, 0)

    pushes_int, int_value = is_int_push_ins(index_stack_value.instruction)
    if pushes_int and isinstance(int_value, int):
        # index_stack_value is an integer and the index is known to the tool
        return TransactionIndex(IndexType.Absolute, int_value)

    if pushes_int and not isinstance(int_value, int):
        # pushes an integer but we don't its value. Unknown index.
        return TransactionIndex(IndexType.Unknown, 0)

    # does not push an literal int value.
    # check for relative index patterns.
    # Txn.group_index() - Int(n)
    if isinstance(index_stack_value.instruction, Sub):
        arg1, arg2 = index_stack_value.args[0], index_stack_value.args[1]
        # index value is arg1 - arg2.
        if isinstance(arg1, UnknownStackValue) or isinstance(arg2, UnknownStackValue):
            return TransactionIndex(IndexType.Unknown, 0)
        if not (
            isinstance(arg1.instruction, Txn) and isinstance(arg1.instruction.field, GroupIndex)
        ):
            return TransactionIndex(IndexType.Unknown, 0)

        # first argument is Txn.group_index()
        is_int, int_value = is_int_push_ins(arg2.instruction)
        if is_int and isinstance(int_value, int):
            offset = -int_value  # pylint: disable=invalid-unary-operand-type
            return TransactionIndex(IndexType.Relative, offset)
        return TransactionIndex(IndexType.Unknown, 0)

    # Txn.group_index() + Int(n)
    if isinstance(index_stack_value.instruction, Add):
        arg1, arg2 = index_stack_value.args[0], index_stack_value.args[1]
        # index value is arg1 + arg2.
        if isinstance(arg1, UnknownStackValue) or isinstance(arg2, UnknownStackValue):
            return TransactionIndex(IndexType.Unknown, 0)
        offset_arg = arg2
        if not (
            isinstance(arg1.instruction, Txn) and isinstance(arg1.instruction.field, GroupIndex)
        ):
            if not (
                isinstance(arg2.instruction, Txn) and isinstance(arg2.instruction.field, GroupIndex)
            ):
                return TransactionIndex(IndexType.Unknown, 0)
            # Int(n) + Txn.group_index()
            offset_arg = arg1

        is_int, int_value = is_int_push_ins(offset_arg.instruction)
        if is_int and isinstance(int_value, int):
            offset = int_value
            return TransactionIndex(IndexType.Relative, offset)
        return TransactionIndex(IndexType.Unknown, 0)

    return TransactionIndex(IndexType.Unknown, 0)



"""Parser for teal contracts.

A TEAL contract is represented using Teal class in tealer. Teal class
comes with methods to register, run analyzers(detectors) and
printers on the parsed contract. Parsing a teal contract is
creating the Control Flow Graph(CFG) of the teal contract,
constructing Teal class instance with the CFG, subroutines and
other related information.

Additional to parsing the teal contracts, subroutines defined in
the contract are identified if there are any, versions of instructions
are verified against the contract version and also attempts to detect mode
of the contract. All the information is stored in the Teal class
representing the contract.

Contracts are parsed in four passes:

#. Parses instructions and adds sequential instruction links.
#. Adds jump based instruction links.
#. Constructs basic blocks and adds sequential basic block links.
#. Adds jump based basic block links.

The final result of parsing is the Control Flow Graph(CFG) of the
contract represented by sequence of the basic blocks.

"""
# pylint: disable=too-many-lines

import sys
import logging
from typing import Optional, Dict, List, Tuple
from collections import defaultdict

from tealer.teal.basic_blocks import BasicBlock
from tealer.teal.subroutine import Subroutine
from tealer.teal.instructions.instructions import (
    Instruction,
    Label,
    B,
    Err,
    BNZ,
    BZ,
    Return,
    Callsub,
    Retsub,
    Pragma,
    Switch,
    Match,
    Intcblock,
    Bytecblock,
    Txn,
    Method,
)

from tealer.teal.instructions.parse_instruction import parse_line, ParseError
from tealer.teal.instructions.transaction_field import TransactionField, ApplicationID
from tealer.teal.instructions.asset_holding_field import AssetHoldingField
from tealer.teal.instructions.asset_params_field import AssetParamsField
from tealer.teal.instructions.app_params_field import AppParamsField
from tealer.teal.instructions.acct_params_field import AcctParamsField
from tealer.teal.teal import Teal
from tealer.utils.arc4_abi import get_method_selector
from tealer.utils.teal_enums import ExecutionMode


logger_parsing = logging.getLogger("Parsing")
logging.basicConfig()


def _detect_execution_mode(instructions: List[Instruction]) -> ExecutionMode:
    """Determine type of contract given its instructions.

    If any of the instructions is supported in only one execution mode then return that as
    contract's execution mode. Otherwise, return ExecutionMode.ANY

    Args:
        instructions: List of all the instructions present in the contract.

    Returns:
        Returns the execution mode of the contract: Stateful, Stateless or Any.
    """

    for ins in instructions:
        if ins.mode != ExecutionMode.ANY:
            return ins.mode
    return ExecutionMode.ANY


def create_bb(instructions: List[Instruction], all_bbs: List[BasicBlock]) -> None:
    """Construct basic blocks and add basic block default edges.

    Third pass of the teal parser. Instructions are
    divided into basic blocks based on the following rules:

    * Label instruction is destination for multiple instructions. A new
        basic block is created for each Label instruction.
    * Instructions B, Err, Return, Callsub, Retsub are exit instructions
        of a basic block. A new basic block is created for instructions
        following them.
    * A instruction with multiple next instructions is a exit instruction
        of the basic block. A new basic block is created for instructions
        following them.

    This function also adds default edges between the constructed basic
    blocks. Two basic blocks have default edge if the entry instruction
    of the second basic block is default next instruction for the exit instruction
    of the first basic block.

    Args:
        instructions: List of instructions.
        all_bbs: (in-place) List of BasicBlocks representing the teal contract.
    """

    bb = BasicBlock()
    all_bbs.append(bb)
    for ins in instructions:

        # if instruction is label and basic block is not empty:
        #   Create a new basic block and add default edge
        # if basic block is empty, no need to create new one.
        if isinstance(ins, Label) and len(bb.instructions) != 0:
            next_bb = BasicBlock()
            all_bbs.append(next_bb)
            bb.add_next(next_bb)
            next_bb.add_prev(bb)
            bb = next_bb

        bb.add_instruction(ins)
        ins.bb = bb

        # `BZ`, `BNZ`, `match` and `switch` have more than one next instruction.
        # `Callsub` has only one next instruction and it is the default next instruction
        # For these instructions, execution continues on the next instruction by default.
        if len(ins.next) > 1 or isinstance(ins, Callsub):
            # if the instruction is the last instruction, do not create a new *empty* block.
            if ins == instructions[-1]:
                continue
            next_bb = BasicBlock()
            all_bbs.append(next_bb)
            # add sequential link
            bb.add_next(next_bb)
            next_bb.add_prev(bb)
            bb = next_bb

        # Err, Return, Retsub will have `len(ins.next) == 0` and `B`.
        if len(ins.next) == 0 or isinstance(ins, B):
            # if the instruction is the last instruction, do not create a *empty* block.
            if ins == instructions[-1]:
                continue
            next_bb = BasicBlock()
            all_bbs.append(next_bb)
            # Do not add any edges. There are no default edges and Jump edges are added in the next pass.
            bb = next_bb


def _add_instruction_comments(ins: Instruction) -> None:
    if isinstance(ins, Txn) and isinstance(ins.field, ApplicationID):
        ins.tealer_comments.append("ApplicationID is 0 in Creation Txn")
    elif isinstance(ins, Method):
        method_signature = ins.method_signature
        method_signature.strip('"')  # quotes are not removed while parsing
        method_selector = get_method_selector(method_signature)
        ins.tealer_comments.append(f"method-selector: {method_selector}")


def first_pass(  # pylint: disable=too-many-branches
    lines: List[str],
    labels: Dict[str, Label],
    subroutines: Dict[str, List[Callsub]],
    instructions: List[Instruction],
) -> Tuple[List[Intcblock], List[Bytecblock]]:
    """Parse instructions and add default instruction edges.

    First pass of the teal parser. Source code lines are parsed into corresponding Instruction objects and default
    edges are added to the instructions.
    * default edges are edges between two continuous instructions where execution might pass from the first
        instruction to the second instruction.
    * Jump edges are edges between two non-continuous instructions where execution might pass from the first instruction
        to the second.

    Args:
        lines: List of source code lines of a teal contract.
        labels: (in-place) Map from teal label string to the parsed label instruction.
            instance.
        subroutines: (in-place) Map from subroutine name to callsub instructions calling the subroutine.
        instructions: (in-place) List of parsed instructions.

    Returns:
        intcblock_ins: List of `intcblock` instructions in the contract.
        bytecblock_ins: List of `bytecblock` instructions in the contract.
    """

    # First pass over the intructions list: Add default edges and collect other information.
    idx = 0
    prev: Optional[Instruction] = None  # Flag: last instruction was an unconditional jump
    intcblock_ins: List[Intcblock] = []  # List of intcblock instructions present in the contract
    bytecblock_ins: List[Bytecblock] = []  # List of bytecblock instructions present in the contract

    instruction_comments: List[str] = []
    for line in lines:
        try:
            if line.strip().startswith("//"):
                # is a comment without any instruction
                instruction_comments.append(line)
                ins = None
            else:
                ins = parse_line(line)
                if ins and instruction_comments:
                    ins.comments_before_ins = list(instruction_comments)
                    instruction_comments = []
        except ParseError as e:
            print(f"Parse error at line {idx}: {e}")
            sys.exit(1)
        idx = idx + 1
        if not ins:
            continue

        if isinstance(ins, Intcblock):
            intcblock_ins.append(ins)
        elif isinstance(ins, Bytecblock):
            bytecblock_ins.append(ins)

        ins.line = idx
        _add_instruction_comments(ins)

        # A label? Add it to the global label list
        if isinstance(ins, Label):
            labels[ins.label] = ins

        # If the prev. ins. was anything **other** than an unconditional jump or unconditional exit:
        #   then add default between the two instructions
        if prev:
            ins.add_prev(prev)
            prev.add_next(ins)

        # A flag that says that this was an unconditional jump or unconditional exit instructions.
        # `Callsub`` is not an unconditional jump instruction.
        # `B` is unconditional jump.
        # `Err, Return` is unconditional exit of the program.
        # `Retsub` is unconditional exit of the subroutine.
        prev = ins
        if isinstance(ins, (B, Err, Return, Retsub)):
            prev = None

        # if ins is callsub, add the label to subroutines and store callsub instruction.
        if isinstance(ins, Callsub):
            subroutines[ins.label].append(ins)

        # Finally, add the instruction to the instruction list
        instructions.append(ins)
    return intcblock_ins, bytecblock_ins


def second_pass(  # pylint: disable=too-many-branches
    instructions: List[Instruction],
    labels: Dict[str, Label],
) -> None:
    """Add jump edges between instructions.

    Second pass of the teal parser.
    * Jump edges are edges between two non-continuous instructions where execution might pass from the first instruction
        to the second.
    * Jump instructions: `B`, `BZ`, `BNZ`, `match`, `switch`.

    Args:
        instructions: List of instructions.
        labels: Map from teal label string to the parsed label instruction.
    """
    logger_parsing.debug("Second Pass")
    # Second pass over the instructions list: Add instruction links for jumps
    for ins in instructions:

        # If a labeled jump to a single instruction, link the ins to its label
        if isinstance(ins, (B, BZ, BNZ)):
            ins.add_next(labels[ins.label])
            labels[ins.label].add_prev(ins)

        # if switch or match, link the ins to its labels
        if isinstance(ins, (Switch, Match)):
            for ins_label in ins.labels:
                ins.add_next(labels[ins_label])
                labels[ins_label].add_prev(ins)


def fourth_pass(basic_blocks: List[BasicBlock]) -> None:  # pylint: disable=too-many-branches
    """Add jump edges between basic blocks.

    Fourth pass of the teal parser. Jump edges
    are added between two basic blocks if there is a jump edge from
    exit instruction of the first basic block to the entry instruction
    of the second basic block.

    Args:
        basic_blocks: List of basic blocks.
    """

    # Fourth pass over the basic blocks: Add jump edges between basic blocks
    for bb in basic_blocks:
        ins = bb.exit_instr
        for next_ins in ins.next:
            next_bb = next_ins.bb
            if next_bb not in bb.next:
                assert bb not in next_bb.prev
                bb.add_next(next_bb)
                next_bb.add_prev(bb)


def _add_basic_blocks_idx(bbs: List[BasicBlock]) -> List[BasicBlock]:
    """Set index of basic blocks based on their entry instruction line number.

    Basic Blocks are ordered in the increasing order based on the
    line number of their entry instructions. Index starts from 0.

    Args:
        bbs: List of BasicBlock objects representing the teal contract.

    Returns:
        Returns :bbs: after sorting and updating their indexes(idx).
    """

    bbs = sorted(bbs, key=lambda x: x.entry_instr.line)
    for i, bb in enumerate(bbs):
        bb.idx = i
    return bbs


def identify_subroutine_blocks(entry_block: "BasicBlock") -> List["BasicBlock"]:
    """find all the basic blocks part of a subroutine using DFS.

    Args:
        entry_block: Entry block of the subroutine.

    Returns:
        Returns the list of all basic blocks part of a subroutine.
    """

    subroutines_blocks: List["BasicBlock"] = []
    stack: List["BasicBlock"] = []

    stack.append(entry_block)
    while len(stack) > 0:
        bb = stack.pop()
        subroutines_blocks.append(bb)

        for next_bb in bb.next:
            if next_bb not in subroutines_blocks and next_bb not in stack:
                stack.append(next_bb)

    return subroutines_blocks


def _verify_version(ins_list: List[Instruction], program_version: int) -> bool:
    """Verify contract instructions and fields are supported in given teal version.

    verify that instructions and fields used in the contract are supported in the
    Teal version specified by the contract. This function doesn't raise an exception
    in case of error, only returns a boolean representing the presence of it and
    prints related information to stderr.

    Args:
        ins_list (List[Instruction]): list of contract instructions.
        program_version (int): Teal version to check against the instruction and
            fields version.

    Returns:
        (bool): returns true if there is any error i.e if any of the instructions or fields are
        not supported in the contract version Or if the contract contains instructions that are
        specific to both Signature and Application Mode.
    """

    stateful_ins: List[Instruction] = []
    stateless_ins: List[Instruction] = []
    error = False

    for ins in ins_list:
        if program_version < ins.version:
            print(
                f"{ins.line}: {ins} instruction is not supported in Teal version {program_version}"
                f", it is supported from Teal version {ins.version}",
                file=sys.stderr,
            )
            error = True
        else:
            field = getattr(ins, "field", None)
            if field is not None and isinstance(
                field,
                (
                    TransactionField,
                    AssetHoldingField,
                    AssetParamsField,
                    AppParamsField,
                    AcctParamsField,
                ),
            ):
                if program_version < field.version:
                    print(
                        f"{ins.line}: {ins}, field {field} is not supported in Teal version {program_version}"
                        f", it is supported from Teal version {field.version}",
                        file=sys.stderr,
                    )
                    error = True
        if ins.mode == ExecutionMode.STATEFUL:
            stateful_ins.append(ins)
        elif ins.mode == ExecutionMode.STATELESS:
            stateless_ins.append(ins)

    if stateless_ins and stateful_ins:
        print(
            "\nprogram contains instructions specific to both Application and Signature Mode",
            file=sys.stderr,
        )
        print("Instructions supported only in Signature Mode:", file=sys.stderr)
        for ins in stateless_ins:
            print(f"\t{ins.line}: {ins}", file=sys.stderr)
        print("\nInstructions supported only in Application Mode:", file=sys.stderr)
        for ins in stateful_ins:
            print(f"\t{ins.line}: {ins}", file=sys.stderr)
        error = True

    return error


def _fill_intc_bytec_info(
    intcblock_ins: List[Intcblock],
    bytecblock_ins: List[Bytecblock],
    entry_block: BasicBlock,
    teal: Teal,
) -> None:
    """Find intcblock, bytecblock instructions and fill that information in Teal class

    Tealer determines intc_*/bytec_* instruction values if and only if intcblock, bytecblock
    instructions are used only once and that too in the entry block itself.

    This is the case for most of the contracts as intcblock/bytecblock instructions are generated
    internally by the assembler or PyTeal compiler.

    Args:
        intcblock_ins: List of all "intcblock" instructions in the contract.
        bytecblock_ins: List of all "bytecblock" instructions in the contract.
        entry_block: Entry block of the contract.
        teal: The contract.
    """
    if len(intcblock_ins) == 1 and intcblock_ins[0].bb == entry_block:
        teal.set_int_constants(intcblock_ins[0].constants)

    if len(bytecblock_ins) == 1 and bytecblock_ins[0].bb == entry_block:
        teal.set_byte_constants(bytecblock_ins[0].constants)


def parse_teal(  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    source_code: str, contract_name: str = ""
) -> Teal:
    """Parse algorand smart contracts written in teal.

    Parsing teal cotracts consists of four passes:

    #. Parses instructions and adds default edges between instructions.
    #. Adds jump edges between links.
    #. Constructs basic blocks and adds default edges between basic block.
    #. Adds jump edges between basic blocks.

    Args:
        source_code: TEAL source code of the contract.
        contract_name: Name of the contract.

    Returns:
        Teal representing the given contract.
    """

    instructions: List[Instruction] = []  # Parsed instructions list
    labels: Dict[str, Label] = {}  # Global map of label names to label instructions
    # Map from subroutine name to callsub instructions calling it
    subroutine_callsubs: Dict[str, List[Callsub]] = defaultdict(list)

    lines = source_code.splitlines()

    intcblock_ins, bytecblock_ins = first_pass(lines, labels, subroutine_callsubs, instructions)
    logger_parsing.debug(f"subroutine_callsubs = {subroutine_callsubs}")
    second_pass(instructions, labels)
    logger_parsing.debug("instruction and nexts")
    for ins in instructions:
        logger_parsing.debug(f"     {ins}, next: {ins.next}")

    # Third pass over the instructions list: Construct the basic blocks and sequential links
    all_bbs: List[BasicBlock] = []
    create_bb(instructions, all_bbs)

    fourth_pass(all_bbs)

    all_bbs = _add_basic_blocks_idx(all_bbs)
    mode = _detect_execution_mode(instructions)

    version = 1
    if isinstance(instructions[0], Pragma):
        version = instructions[0].program_version

    _verify_version(instructions, version)

    all_reachable_blocks: List["BasicBlock"] = []
    subroutines_and_blocks: Dict[str, List["BasicBlock"]] = {}
    for subroutine_name in subroutine_callsubs:
        subroutine_entry_block = labels[subroutine_name].bb
        subroutine_blocks = identify_subroutine_blocks(subroutine_entry_block)
        subroutines_and_blocks[subroutine_name] = subroutine_blocks
        all_reachable_blocks += subroutine_blocks

    contract_entry_block = all_bbs[0]
    main_entry_point_blocks = identify_subroutine_blocks(contract_entry_block)
    all_reachable_blocks += main_entry_point_blocks

    subroutines: Dict[str, "Subroutine"] = {}
    for subroutine_name in subroutine_callsubs:
        label_ins = labels[subroutine_name]
        subroutine_entry_block = label_ins.bb
        # add tealer comment "Subroutine: {label}" to the subroutine entry block
        subroutine_entry_block.tealer_comments.append(f"Subroutine {subroutine_name}")
        # list all blocks of the subroutine using DFS
        subroutine_blocks = subroutines_and_blocks[subroutine_name]
        subroutine_obj = Subroutine(subroutine_name, subroutine_entry_block, subroutine_blocks)
        # set callsub blocks calling the subroutine
        callsub_blocks = [ins.bb for ins in subroutine_callsubs[subroutine_name]]
        callsub_blocks = [bb for bb in callsub_blocks if bb in all_reachable_blocks]
        subroutine_obj.caller_blocks = callsub_blocks
        # for each callsub instruction, set the called subroutine
        for ins in subroutine_callsubs[subroutine_name]:
            ins.called_subroutine = subroutine_obj

        # set subroutine to each basic block
        for bi in subroutine_blocks:
            bi.subroutine = subroutine_obj
        subroutines[subroutine_name] = subroutine_obj

    main_program_name = "__main__"
    main_program = Subroutine(main_program_name, contract_entry_block, main_entry_point_blocks)
    for bi in main_entry_point_blocks:
        bi.subroutine = main_program

    # TODO: Handle unreachable basic blocks.
    # Note: PyTeal generated contracts have unreachable code.
    for bi in all_bbs:
        if bi not in all_reachable_blocks:
            # bi is unreachable
            for bnext in bi.next:
                bnext.prev.remove(bi)
                bi.next.remove(bnext)

            for ins_next in bi.exit_instr.next:
                ins_next.prev.remove(bi.exit_instr)
                bi.exit_instr.next.remove(ins_next)

            for ins in bi.instructions:
                instructions.remove(ins)

    all_reachable_blocks = sorted(list(set(all_reachable_blocks)), key=lambda bi: bi.idx)
    teal = Teal(version, mode, instructions, all_reachable_blocks, main_program, subroutines)

    # set teal instance for it's basic blocks
    for bb in teal.bbs:
        bb.teal = teal
        bb.tealer_comments.insert(0, f"block_id = {bb.idx}; cost = {bb.cost}")

    for subroutine in [main_program] + list(subroutines.values()):
        subroutine.contract = teal

    teal.contract_name = contract_name
    _fill_intc_bytec_info(intcblock_ins, bytecblock_ins, teal.main.entry, teal)

    return teal



"""Defines classes to represent global fields.

Global fields are fields that are common to all the transactions
in the group. Teal supports access to this fields using ``global``
instruction.

Each global field is represented by a class in tealer. All the
classes representing the fields must inherit from GlobalField class.

"""

# pylint: disable=too-few-public-methods
class GlobalField:
    """Base class representing a global field"""

    def __init__(self) -> None:
        self._version: int = 1

    @property
    def version(self) -> int:
        """Teal version this field is introduced in and supported from.

        Returns:
            Teal version the field is introduced in and supported from.
        """
        return self._version

    def __str__(self) -> str:
        return self.__class__.__qualname__


class GroupSize(GlobalField):
    """Number of transactions in this atomic transaction group."""


class MinTxnFee(GlobalField):
    """Minimum transaction fee in micro algos."""


class ZeroAddress(GlobalField):
    """32 byte address of all zero bytes"""


class MinBalance(GlobalField):
    """Minimum balance in micro algos."""


class MaxTxnLife(GlobalField):
    """Maximum Transaction Life in number of rounds."""


class LogicSigVersion(GlobalField):
    """Maximum supported TEAL version."""

    def __init__(self) -> None:
        super().__init__()
        self._version = 2


class Round(GlobalField):
    """Current round number."""

    def __init__(self) -> None:
        super().__init__()
        self._version = 2


class LatestTimestamp(GlobalField):
    """Last confirmed block unix timestamp. Fails if negative."""

    def __init__(self) -> None:
        super().__init__()
        self._version = 2


class CurrentApplicationID(GlobalField):
    """ID of the current application executing."""

    def __init__(self) -> None:
        super().__init__()
        self._version = 2


class CreatorAddress(GlobalField):
    """Address of the creator of the current application."""

    def __init__(self) -> None:
        super().__init__()
        self._version = 3


class CurrentApplicationAddress(GlobalField):
    """Address that the current application controls."""

    def __init__(self) -> None:
        super().__init__()
        self._version = 5


class GroupID(GlobalField):
    """ID of the transaction group.

    ID will be 32 zero bytes if the transaction is not part
    of the group.
    """

    def __init__(self) -> None:
        super().__init__()
        self._version = 5


class OpcodeBudget(GlobalField):
    """The remaining cost that can be spent by opcodes in the program"""

    def __init__(self) -> None:
        super().__init__()
        self._version = 6


class CallerApplicationID(GlobalField):
    """The application ID of the application that called this application.

    ID wil be 0 if the current application is at the top-level. Application mode only.
    """

    def __init__(self) -> None:
        super().__init__()
        self._version = 6


class CallerApplicationAddress(GlobalField):
    """The application address of the application that called this application.

    ZeroAddress if this application is at the top-level. Application mode only.
    """

    def __init__(self) -> None:
        super().__init__()
        self._version = 6



"""Defines classes to represent acct_params_get fields.

``acct_params_get`` instruction is used to access fields
of an account in the contract.

Each field that can be accessed using acct_params_get is represented
by a class in tealer. All the classes representing the fields must
inherit from AcctParamsField class.

"""

# pylint: disable=too-few-public-methods
class AcctParamsField:
    """Base class to represent Acct fields."""

    def __init__(self) -> None:
        self._version: int = 6

    @property
    def version(self) -> int:
        """Teal version this field is introduced in and supported from.

        Returns:
            Teal version the field is introduced in and supported from.
        """
        return self._version

    def __str__(self) -> str:
        return self.__class__.__qualname__


class AcctBalance(AcctParamsField):
    """Account balance in microalgos."""


class AcctMinBalance(AcctParamsField):
    """Minimum required balance for account, in microalgos."""


class AcctAuthAddr(AcctParamsField):
    """Address the account is rekeyed to."""


class AcctTotalNumUint(AcctParamsField):
    """The total number of uint64 values allocated by this account in Global and Local States."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8


class AcctTotalNumByteSlice(AcctParamsField):
    """The total number of byte array values allocated by this account in Global and Local States."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8


class AcctTotalExtraAppPages(AcctParamsField):
    """The number of extra app code pages used by this account."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8


class AcctTotalAppsCreated(AcctParamsField):
    """The number of existing apps created by this account"""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8


class AcctTotalAppsOptedIn(AcctParamsField):
    """The number of apps this account is opted into."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8


class AcctTotalAssetsCreated(AcctParamsField):
    """The number of existing ASAs created by this account."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8


class AcctTotalAssets(AcctParamsField):
    """The numbers of ASAs held by this account (including ASAs this account created)."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8


class AcctTotalBoxes(AcctParamsField):
    """The number of existing boxes created by this account's app."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8


class AcctTotalBoxBytes(AcctParamsField):
    """The total number of bytes used by this account's app's box keys and values."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 8



from typing import Dict, Any, List
from dataclasses import dataclass
from pathlib import Path
import yaml


class InvalidTestCase(Exception):
    pass


@dataclass
class OutputContract:
    contract_name: str
    function_name: str

    @staticmethod
    def from_yaml(contract: Dict[str, Any]) -> "OutputContract":
        if "contract" not in contract or "function" not in contract:
            raise InvalidTestCase(
                f"Missing required fields in the contract description: {contract}"
            )
        return OutputContract(contract["contract"], contract["function"])


@dataclass
class OutputTransaction:
    transaction_id: str
    contracts: List[OutputContract]

    @staticmethod
    def from_yaml(transaction: Dict[str, Any]) -> "OutputTransaction":
        if "txn_id" not in transaction:
            raise InvalidTestCase(f'Missing "txn_id" field for the transaction {transaction}')

        contracts = []
        if "contracts" in transaction:
            for contract in transaction["contracts"]:
                contracts.append(OutputContract.from_yaml(contract))
        return OutputTransaction(
            transaction["txn_id"],
            sorted(contracts, key=lambda x: (x.contract_name, x.function_name)),
        )


@dataclass
class OutputOperation:
    operation: str
    vulnerable_transactions: List[OutputTransaction]

    @staticmethod
    def from_yaml(operation: Dict[str, Any]) -> "OutputOperation":
        if "operation" not in operation or "vulnerable_transactions" not in operation:
            raise InvalidTestCase(f"Missing required fields in the operation: {operation}")

        operation_name = operation["operation"]

        vulnerable_transactions = []
        for transaction in operation["vulnerable_transactions"]:
            vulnerable_transactions.append(OutputTransaction.from_yaml(transaction))

        return OutputOperation(operation_name, vulnerable_transactions)


@dataclass
class ExpectedOutput:
    name: str
    operations: List[OutputOperation]

    @staticmethod
    def from_yaml(output: Dict[str, Any]) -> "ExpectedOutput":
        if "name" not in output or "operations" not in output:
            raise InvalidTestCase(f"Missing required fields in the expected output: {output}")

        name = output["name"]
        operations = []
        for operation in output["operations"]:
            operations.append(OutputOperation.from_yaml(operation))
        return ExpectedOutput(name, operations)


def expected_output_from_file(file_path: Path) -> "ExpectedOutput":
    with open(file_path, encoding="utf-8") as f:
        # print(yaml.safe_load(f.read()))
        return ExpectedOutput.from_yaml(yaml.safe_load(f.read()))



# type: ignore[unreachable]
import sys
import pytest

from tealer.teal.instructions import instructions
from tealer.teal.parse_teal import parse_teal

if not sys.version_info >= (3, 10):
    pytest.skip(reason="PyTeal based tests require python >= 3.10", allow_module_level=True)

# pylint: disable=wrong-import-position
# Place import statements after the version check
from tests.pyteal_parsing.normal_application import normal_application_approval_program
from tests.pyteal_parsing.arc4_application import arc4_application_ap
from tests.pyteal_parsing.control_flow_constructs import control_flow_ap


TARGETS = [
    normal_application_approval_program,
    arc4_application_ap,
    control_flow_ap,
]


@pytest.mark.parametrize("target", TARGETS)  # type: ignore
def test_parsing_using_pyteal(target: str) -> None:
    teal = parse_teal(target)
    # print instruction to trigger __str__ on each ins
    # print stack pop/push to trigger the function on each ins
    for i in teal.instructions:
        assert not isinstance(i, instructions.UnsupportedInstruction), f'ins "{i}" is not supported'
        print(i, i.cost)
        print(i, i.stack_pop_size)
        print(i, i.stack_push_size)



import re
import pytest

from tealer.teal.parse_teal import parse_teal
from tealer.teal.instructions.transaction_field import TransactionField
from tealer.utils.teal_enums import ExecutionMode


INSTRUCTIONS_TARGETS = [
    "tests/parsing/teal-instructions-with-versions.teal",
]

FIELDS_TARGETS = [
    "tests/parsing/teal-fields-with-versions.teal",
]

instruction_info_pattern = re.compile(
    r"\(version: ([0-9]+), mode: (Any|Stateful|Stateless), pop: ([0-9]+), push: ([0-9+])\)"
)

field_info_pattern = re.compile(r"\(version: ([0-9]+)\)")


@pytest.mark.parametrize("target", INSTRUCTIONS_TARGETS)  # type: ignore
def test_instruction_version(target: str) -> None:
    with open(target, encoding="utf-8") as f:
        teal = parse_teal(f.read())
    for ins in teal.instructions:
        match_obj = instruction_info_pattern.search(ins.comment)
        if match_obj is None:
            continue
        version, mode_str, pop_size, push_size = match_obj.groups()
        mode = {
            "Any": ExecutionMode.ANY,
            "Stateful": ExecutionMode.STATEFUL,
            "Stateless": ExecutionMode.STATELESS,
        }[mode_str]
        assert ins.mode == mode
        assert ins.version == int(version)
        assert ins.stack_push_size == int(push_size)
        assert ins.stack_pop_size == int(pop_size)


@pytest.mark.parametrize("target", FIELDS_TARGETS)  # type: ignore
def test_field_version(target: str) -> None:
    with open(target, encoding="utf-8") as f:
        teal = parse_teal(f.read())
    for ins in teal.instructions:
        field = getattr(ins, "field", None)
        if field is None:
            continue
        if isinstance(field, TransactionField):
            match_obj = field_info_pattern.search(ins.comment)
            if match_obj is None:
                continue
            version = match_obj.groups()[0]
            assert field.version == int(version)



# pylint: disable=unused-import
from tealer.analyses.dataflow.transaction_context.int_fields import GroupIndices
from tealer.analyses.dataflow.transaction_context.txn_types import TxnType
from tealer.analyses.dataflow.transaction_context.addr_fields import AddrFields
from tealer.analyses.dataflow.transaction_context.fee_field import FeeField
